# COPYRIGHT 2011 Brian House
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this
# software and associated documentation files (the “Software”), to deal in the Software
# without restriction, including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
# FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
"""
Euclidean pattern generating library

Copied from https://github.com/brianhouse/bjorklund with all due gratitude

Originally written by Brian House (c) 2011. Released under the MIT license
"""


def generate_euclidean_pattern(steps, pulses, rot=0):
    """Generates an array indicating the on/off steps of Euclid(k, n)

    :param steps:  The number of steps in the pattern
    :param pulses: The number of ON steps in the pattern (must be <= steps)
    :param rot:    Optional rotation to offset the pattern. Must be in the range [0, steps]

    :return: An int array of length steps consisting of 1 and 0 values only

    :raises ValueError: if pulses or rot is out of range
    """
    steps = int(steps)
    pulses = int(pulses)
    rot = int(rot)
    if pulses > steps:
        raise ValueError("Pulses cannot be greater than steps")
    if pulses < 0:
        raise ValueError("Pulses must be positive")
    if rot > steps:
        raise ValueError("Rotation cannot be greater than steps")
    if steps < 0:
        raise ValueError("Steps must be positive")
    if steps == 0:
        return []
    if pulses == 0:
        return [0] * steps
    pattern = []
    counts = []
    remainders = []
    divisor = steps - pulses
    remainders.append(pulses)
    level = 0
    while True:
        counts.append(divisor // remainders[level])
        remainders.append(divisor % remainders[level])
        divisor = remainders[level]
        level = level + 1
        if remainders[level] <= 1:
            break
    counts.append(divisor)

    def build(level):
        if level == -1:
            pattern.append(0)
        elif level == -2:
            pattern.append(1)
        else:
            for i in range(0, counts[level]):
                build(level - 1)
            if remainders[level] != 0:
                build(level - 2)

    build(level)
    i = pattern.index(1)
    pattern = pattern[i:] + pattern[0:i]

    # rotate the pattern if needed by removing the last item and
    # adding it to the start
    if rot != 0:
        for i in range(rot):
            x = pattern.pop(-1)
            pattern.insert(0, x)

    return pattern



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import os
import json

from europi_log import *


def load_file(filename, mode: str = "r") -> object:
    """Load a file and return its contents

    :param filename:  The name of the file to load
    :param mode:      The mode to open the file in. Should be "r" for text files or "rb" for binary files

    :return:  The file's contents, either as a string or bytes, depending on mode.
    """
    try:
        with open(filename, mode) as file:
            return file.read()
    except OSError as e:
        log_warning(f"Unable to read {filename}: {e}", "file_utils")
        if "b" in mode:
            return b""
        else:
            return ""


def load_json_file(filename, mode="r") -> dict:
    """Load a file and return its contents

    :param filename:  The name of the file to load
    :param mode:      The mode to open the file in. Should be "r" except in very unique circumstances

    :return:  The file's contents as a dict
    """
    try:
        with open(filename, mode) as file:
            return json.load(file)
    except ValueError as e:
        log_warning(f"Unable to parse JSON data from {filename}: {e}", "file_utils")
        return {}
    except OSError as e:
        if e.errno == errno.ENOENT:
            log_info(f"/{filename} does not exist. Using default settings", "file_utils")
        else:
            log_warning(f"Unable to open {filename}: {e}", "file_utils")
        return {}


def delete_file(filename):
    """
    Delete a file from the disk if it exists

    :param filename:  The file to delete
    """
    try:
        os.remove(filename)
    except OSError:
        pass



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from machine import ADC
from time import sleep

from europi import (
    OLED_HEIGHT,
    OLED_WIDTH,
    ain,
    b1,
    b2,
    cv1,
    cv2,
    cv3,
    cv4,
    cv5,
    cv6,
    din,
    k1,
    k2,
    oled,
    europi_config,
    thermometer,
)
from europi_script import EuroPiScript
import configuration

"""
A diagnostic utility intended to help prove out a new EuroPi build and calibration. Each aspect of the EuroPi's hardware
is exercised.

- din: value displayed on screen
- ain: value displayed on screen
- b1: rotate output voltages backwards
- b2: rotate output voltages forwards
- k1: value 0-99 displayed on screen
- k2: value 0-99 displayed on screen
- cvX: output a constant voltage, one of [0, 0.5, 1, 2.5, 5, 10]
"""


class Diagnostic(EuroPiScript):
    def __init__(self):
        super().__init__()
        self.voltages = [
            0,  # min
            0.5,  # not 0 but still below DI's threshold
            1,
            2.5,
            5,
            10,  # max
        ]
        self.temp_units = self.config.TEMP_UNITS
        self.use_fahrenheit = self.temp_units == "F"

    @classmethod
    def config_points(cls):
        return [configuration.choice(name="TEMP_UNITS", choices=["C", "F"], default="C")]

    def calc_temp(self):
        t = thermometer.read_temperature()
        if t is None:
            return 0

        if self.use_fahrenheit:
            t = (t * 1.8) + 32
        return t

    def rotate_r(self):
        self.voltages = self.voltages[-1:] + self.voltages[:-1]

    def rotate_l(self):
        self.voltages = self.voltages[1:] + self.voltages[:1]

    def main(self):
        b1.handler(self.rotate_l)
        b2.handler(self.rotate_r)

        while True:
            # Set the outputs to useful values
            cv1.voltage(self.voltages[0])
            cv2.voltage(self.voltages[1])
            cv3.voltage(self.voltages[2])
            cv4.voltage(self.voltages[3])
            cv5.voltage(self.voltages[4])
            cv6.voltage(self.voltages[5])

            oled.fill(0)

            # calc and format temp
            t = self.calc_temp()
            formatted_temp = f"{int(t)}{self.temp_units}"

            # display the input values
            oled.text(f"ain: {ain.read_voltage(samples=512):5.2f}v {formatted_temp}", 2, 3, 1)
            oled.text(f"k1: {k1.read_position():2}  k2: {k2.read_position():2}", 2, 13, 1)
            oled.text(f"din:{din.value()} b1:{b1.value()} b2:{b2.value()}", 2, 23, 1)

            # show the screen boundaries
            oled.rect(0, 0, OLED_WIDTH, OLED_HEIGHT, 1)
            oled.show()

            sleep(0.1)


if __name__ == "__main__":
    Diagnostic().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface class for the DS1307 Realtime Clock

This class is designed to work with a DS1307 chip mounted on an I2C carrier board
that can be connected to EuroPi's external I2C interface. The user is required to
1) provide their own RTC module
2) create/source an appropriate adapter to connect the GND, VCC, SDA, and SCL pins on EuroPi to the RTC module
3) Mount the RTC module securely in such a way that it won't come loose nor accidentally short out any other components.

Based on work by Mike Causer released under the MIT license (c) 2018:
https://github.com/mcauser/micropython-tinyrtc-i2c/blob/master/ds1307.py


NOTE: the author of this module does NOT have access to a DS1307 module; this class is thoroughly
untested. USE AT YOUR OWN RISK.

Hopefully I, or another EuroPi contributor, will get access to the required module to validate the code.
But at present this class is provided as-is, based wholly on Mike Causer's work with the necessary
changes to support EuroPi's RTC interface.
"""

from experimental.clocks.clock_source import ExternalClockSource

try:
    from micropython import const
except ImportError:

    def const(x):
        return x


# fmt: off
DATETIME_REG = const(0)    # 0x00-0x06
CHIP_HALT    = const(128)  # 0x80, 0b10000000
CONTROL_REG  = const(7)    # 0x07
RAM_REG      = const(8)    # 0x08-0x3F
# fmt: on


class DS1307(ExternalClockSource):
    """Driver for the DS1307 RTC."""

    def __init__(self, i2c, addr=0x68):
        super().__init__()
        self.i2c = i2c
        self.addr = addr
        self._halt = False

    def _dec2bcd(self, value):
        """Convert decimal to binary coded decimal (BCD) format"""
        return (value // 10) << 4 | (value % 10)

    def _bcd2dec(self, value):
        """Convert binary coded decimal (BCD) format to decimal"""
        return ((value >> 4) * 10) + (value & 0x0F)

    def datetime(self):
        """
        Get the current time.

        :return: a tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        buf = self.i2c.readfrom_mem(self.addr, DATETIME_REG, 7)
        # fmt: off
        return (
            self._bcd2dec(buf[6]) + 2000,  # year
            self._bcd2dec(buf[5]),         # month
            self._bcd2dec(buf[4]),         # day
            self._bcd2dec(buf[2]),         # hour
            self._bcd2dec(buf[1]),         # minute
            self._bcd2dec(buf[0] & 0x7F),  # second
            self._bcd2dec(buf[3]),         # weekday
            0,                             # yearday (ignored)
        )
        # fmt: on

    def set_datetime(self, datetime):
        """
        Set the current time.

        :param datetime: tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        self.check_valid_datetime(datetime)

        # fmt: off
        buf = bytearray(7)
        try:
            buf[3] = dectobcd(datetime[6])             # Day of week
        except IndexError:
            buf[3] = 0
        try:
            buf[0] = dectobcd(datetime[5])             # Seconds
        except IndexError:
            buf[0] = 0
        buf[1] = dectobcd(datetime[4])                 # Minutes
        buf[2] = dectobcd(datetime[3])                 # Hour + the 24h format flag
        buf[4] = dectobcd(datetime[2])                 # Day
        buf[5] = dectobcd(datetime[1]) & 0xFF          # Month + mask the century flag
        buf[6] = dectobcd(int(str(datetime[0])[-2:]))  # Year can be yyyy, or yy
        if (self._halt):
            buf[0] |= (1 << 7)
        self.i2c.writeto_mem(self.addr, DATETIME_REG, buf)
        # fmt: on

    def halt(self, val=None):
        """Power up, power down or check status"""
        if val is None:
            return self._halt
        reg = self.i2c.readfrom_mem(self.addr, DATETIME_REG, 1)[0]
        if val:
            reg |= CHIP_HALT
        else:
            reg &= ~CHIP_HALT
        self._halt = bool(val)
        self.i2c.writeto_mem(self.addr, DATETIME_REG, bytearray([reg]))

    def square_wave(self, sqw=0, out=0):
        """
        Output square wave on pin SQ

        Available frequencues: 1Hz, 4.096kHz, 8.192kHz or 32.768kHz,
        or disable the oscillator and output logic level high/low.
        """
        rs0 = 1 if sqw == 4 or sqw == 32 else 0
        rs1 = 1 if sqw == 8 or sqw == 32 else 0
        out = 1 if out > 0 else 0
        sqw = 1 if sqw > 0 else 0
        reg = rs0 | rs1 << 1 | sqw << 4 | out << 7
        self.i2c.writeto_mem(self.addr, CONTROL_REG, bytearray([reg]))



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Hardware-specific definitions & implementations for Europi

Originally all of this was included as part of europi.py, but was moved
here to make it easier to access e.g. button status readers independently
of all of the other EuroPi initialization that occurs inside the `europi`
module.

Normally this module should not be included directly; it's imported
automatically as part of `from europi import *`.
"""

from machine import ADC
from machine import I2C
from machine import PWM
from machine import Pin
from machine import freq
from machine import mem32
import time

from europi_config import load_europi_config, CPU_FREQS
from experimental.experimental_config import load_experimental_config

# Load the configuration objects so we can initialize
# the hardware according to user preferences
europi_config = load_europi_config()
experimental_config = load_experimental_config()

# Analogue voltage read range.
MIN_INPUT_VOLTAGE = 0
MAX_INPUT_VOLTAGE = europi_config.MAX_INPUT_VOLTAGE
GATE_VOLTAGE = europi_config.GATE_VOLTAGE
DEFAULT_SAMPLES = 32

# Output voltage range
MIN_OUTPUT_VOLTAGE = 0
MAX_OUTPUT_VOLTAGE = europi_config.MAX_OUTPUT_VOLTAGE

PWM_FREQ = 100_000

# Digital input and output binary values.
HIGH = 1
LOW = 0

# Pin assignments
PIN_DIN = 22
PIN_AIN = 26
PIN_K1 = 27
PIN_K2 = 28
PIN_B1 = 4
PIN_B2 = 5
PIN_CV1 = 21
PIN_CV2 = 20
PIN_CV3 = 16
PIN_CV4 = 17
PIN_CV5 = 18
PIN_CV6 = 19
PIN_USB_CONNECTED = 24
PIN_TEMPERATURE = 4

# Standard max int consts.
MAX_UINT16 = 65535

# How many CV outputs are there?
# On EuroPi this 6, but future versions (e.g. EuroPi X may have more)
NUM_CVS = 6

# Import the calibration values
# These are generated by tools/calibrate.py, but do not exist by default
try:
    from calibration_values import INPUT_CALIBRATION_VALUES, OUTPUT_CALIBRATION_VALUES
except ImportError:
    # Note: run calibrate.py to get a more precise calibration.
    # Default calibration values are close-enough to reasonable performance, but aren't great
    INPUT_CALIBRATION_VALUES = [384, 44634]
    OUTPUT_CALIBRATION_VALUES = [
        0,
        6300,
        12575,
        19150,
        25375,
        31625,
        38150,
        44225,
        50525,
        56950,
        63475,
    ]
# Legacy calibration using only CV1; apply the same calibration values to each output
if type(OUTPUT_CALIBRATION_VALUES[0]) is int:
    cv1_values = OUTPUT_CALIBRATION_VALUES
    OUTPUT_CALIBRATION_VALUES = []
    for i in range(NUM_CVS):
        OUTPUT_CALIBRATION_VALUES.append(cv1_values)


def clamp(value: int | float, low: int | float, high: int | float):
    """
    Returns a value that is no lower than 'low' and no higher than 'high'.

    :param value:  The value to clamp
    :param low:  The inclusive lower bound
    :param high:  The inclusive upper bound
    """
    return max(min(value, high), low)


class AnalogueReader:
    """A base class for common analogue read methods.

    This class in inherited by classes like Knob and AnalogueInput and does
    not need to be used by user scripts.

    :param pin:  The pin ID the ADC is connected to
    :param samples:  The number of ADC samples to read. More samples results in a more
        accurate reading, but may slow down the program
    :param deadzone:  An optional deadzone to restrict noise at the high and low ends
    """

    def __init__(self, pin, samples=DEFAULT_SAMPLES, deadzone=0.0):
        self.pin_id = pin
        self.pin = ADC(Pin(pin))
        self.set_samples(samples)
        self.set_deadzone(deadzone)

    def _sample_adc(self, samples=None):
        # Over-samples the ADC and returns the average.
        value = 0
        for _ in range(samples or self._samples):
            value += self.pin.read_u16()
        return round(value / (samples or self._samples))

    def set_samples(self, samples):
        """Override the default number of sample reads with the given value."""
        if not isinstance(samples, int):
            raise ValueError(f"set_samples expects an int value, got: {samples}")
        self._samples = samples

    def set_deadzone(self, deadzone):
        """Override the default deadzone with the given value."""
        if not isinstance(deadzone, float):
            raise ValueError(f"set_deadzone expects an float value, got: {deadzone}")
        self._deadzone = deadzone

    def percent(self, samples=None, deadzone=None):
        """Return the percentage of the component's current relative range."""
        dz = self._deadzone
        if deadzone is not None:
            dz = deadzone
        value = self._sample_adc(samples) / MAX_UINT16
        value = value * (1.0 + 2.0 * dz) - dz
        return clamp(value, 0.0, 1.0)

    def range(self, steps=100, samples=None, deadzone=None):
        """Return a value (upper bound excluded) chosen by the current voltage value."""
        if not isinstance(steps, int):
            raise ValueError(f"range expects an int value, got: {steps}")
        percent = self.percent(samples, deadzone)
        if int(percent) == 1:
            return steps - 1
        return int(percent * steps)

    def choice(self, values, samples=None, deadzone=None):
        """Return a value from a list chosen by the current voltage value."""
        if not isinstance(values, list):
            raise ValueError(f"choice expects a list, got: {values}")
        percent = self.percent(samples, deadzone)
        if percent == 1.0:
            return values[-1]
        return values[int(percent * len(values))]


class AnalogueInput(AnalogueReader):
    """A class for handling the reading of analogue control voltage.

    The analogue input allows you to 'read' CV from anywhere between 0 and 12V.

    It is protected for the entire Eurorack range, so don't worry about
    plugging in a bipolar source, it will simply be clipped to 0-12V.

    The functions all take an optional parameter of ``samples``, which will
    oversample the ADC and then take an average, which will take more time per
    reading, but will give you a statistically more accurate result. The
    default is 32, provides a balance of performance vs accuracy, but if you
    want to process at the maximum speed you can use as little as 1, and the
    processor won't bog down until you get way up into the thousands if you
    wan't incredibly accurate (but quite slow) readings.

    The percent function takes an optional parameter ``deadzone``. However this
    parameter is ignored and just present to be compatible with the percent
    function of the AnalogueReader and Knob classes

    :param pin:  The pin ID the ADC is connected to
    :param min_voltage:  The minimum allowed input voltage
    :param max_voltage:  The maximum allowed input voltage
    :param deadzone:  An optional deadzone to restrict noise at the high and low ends
    """

    def __init__(self, pin, min_voltage=MIN_INPUT_VOLTAGE, max_voltage=MAX_INPUT_VOLTAGE):
        super().__init__(pin)
        self.MIN_VOLTAGE = min_voltage
        self.MAX_VOLTAGE = max_voltage
        self._gradients = []
        for index, value in enumerate(INPUT_CALIBRATION_VALUES[:-1]):
            try:
                self._gradients.append(1 / (INPUT_CALIBRATION_VALUES[index + 1] - value))
            except ZeroDivisionError:
                raise Exception(
                    "The input calibration process did not complete properly. Please complete again with rack power turned on"
                )
        self._gradients.append(self._gradients[-1])

    def percent(self, samples=None, deadzone=None):
        """Current voltage as a relative percentage of the component's range."""
        # Determine the percent value from the max calibration value.
        reading = self._sample_adc(samples) - INPUT_CALIBRATION_VALUES[0]
        max_value = max(
            reading,
            INPUT_CALIBRATION_VALUES[-1] - INPUT_CALIBRATION_VALUES[0],
        )
        return max(reading / max_value, 0.0)

    def read_voltage(self, samples=None):
        raw_reading = self._sample_adc(samples)
        reading = raw_reading - INPUT_CALIBRATION_VALUES[0]
        max_value = max(
            reading,
            INPUT_CALIBRATION_VALUES[-1] - INPUT_CALIBRATION_VALUES[0],
        )
        percent = max(reading / max_value, 0.0)
        # low precision vs. high precision
        if len(self._gradients) == 2:
            cv = 10 * max(
                reading / (INPUT_CALIBRATION_VALUES[-1] - INPUT_CALIBRATION_VALUES[0]),
                0.0,
            )
        else:
            index = int(percent * (len(INPUT_CALIBRATION_VALUES) - 1))
            cv = index + (self._gradients[index] * (raw_reading - INPUT_CALIBRATION_VALUES[index]))
        return clamp(cv, self.MIN_VOLTAGE, self.MAX_VOLTAGE)


class Knob(AnalogueReader):
    """A class for handling the reading of knob voltage and position.

    Read_position has a default value of 100, meaning if you simply use
    ``kx.read_position()`` you will return a whole number percent style value
    from 0-100.

    There is also the optional parameter of ``samples`` (which must come after the
    normal parameter), the same as the analogue input uses (the knob positions
    are 'read' via an analogue to digital converter). It has a default value
    of 256, but you can use higher or lower depending on if you value speed or
    accuracy more. If you really want to avoid 'noise' which would present as
    a flickering value despite the knob being still, then I'd suggest using
    higher samples (and probably a smaller number to divide the position by).
    The default ``samples`` value can also be set using the ``set_samples()``
    method, which will then be used on all analogue read calls for that
    component.

    An optional ``deadzone`` parameter can be used to place deadzones at both
    positions (all the way left and right) of the knob to make sure the full range
    is available on all builds. The default value is 0.01 (resulting in 1% of the
    travel used as deadzone on each side). There is usually no need to change this.

    Additionally, the ``choice()`` method can be used to select a value from a
    list of values based on the knob's position::

        def clock_division(self):
            return k1.choice([1, 2, 3, 4, 5, 6, 7, 8, 16, 32])

    When the knob is all the way to the left, the return value will be ``1``,
    at 12 o'clock it will return the mid point value of ``5`` and when fully
    clockwise, the last list item of ``32`` will be returned.

    The ADCs used to read the knob position are only 12 bit, which means that
    any read_position value above 4096 (2^12) will not actually be any finer
    resolution, but will instead just go up in steps. For example using 8192
    would only return values which go up in steps of 2.

    :param pin:  The pin ID the ADC is connected to
    :param deadzone:  An optional deadzone to restrict noise at the high and low ends
    """

    def __init__(self, pin, deadzone=0.01):
        super().__init__(pin, deadzone=deadzone)

    def percent(self, samples=None, deadzone=None):
        """Return the knob's position as relative percentage."""
        # Reverse range to provide increasing range.
        return 1.0 - super().percent(samples, deadzone)

    def read_position(self, steps=100, samples=None, deadzone=None):
        """Returns the position as a value between zero and provided integer."""
        return self.range(steps, samples, deadzone)


class DigitalReader:
    """A base class for common digital inputs methods.

    This class in inherited by classes like Button and DigitalInput and does
    not need to be used by user scripts.

    :param pin:  The pin ID we read from
    :param debounce_delay:  The time in ms that we debounce the input
    """

    def __init__(self, pin, debounce_delay=500):
        self.pin = Pin(pin, Pin.IN)
        self.debounce_delay = debounce_delay

        # Default handlers are noop callables.
        self._rising_handler = lambda: None
        self._falling_handler = lambda: None

        # Both high handler
        self._both_handler = lambda: None
        self._other = None

        # IRQ event timestamps
        self.last_rising_ms = 0
        self.last_falling_ms = 0

    def _bounce_wrapper(self, pin):
        """IRQ handler wrapper for falling and rising edge callback functions."""
        if self.value() == HIGH:
            if time.ticks_diff(time.ticks_ms(), self.last_rising_ms) < self.debounce_delay:
                return
            self.last_rising_ms = time.ticks_ms()
            return self._rising_handler()
        else:
            if time.ticks_diff(time.ticks_ms(), self.last_falling_ms) < self.debounce_delay:
                return
            self.last_falling_ms = time.ticks_ms()

            # Check if 'other' pin is set and if 'other' pins is high and if this pin has been high for long enough.
            if (
                self._other
                and self._other.value()
                and time.ticks_diff(self.last_falling_ms, self.last_rising_ms) > 500
            ):
                return self._both_handler()
            return self._falling_handler()

    def value(self):
        """The current binary value, HIGH (1) or LOW (0)."""
        # Both the digital input and buttons are normally high, and 'pulled'
        # low when on, so this is flipped to be more intuitive
        # (high when on, low when off)
        return LOW if self.pin.value() else HIGH

    def handler(self, func):
        """Define the callback function to call when rising edge detected."""
        if not callable(func):
            raise ValueError("Provided handler func is not callable")
        self._rising_handler = func
        self.pin.irq(handler=self._bounce_wrapper)

    def handler_falling(self, func):
        """Define the callback function to call when falling edge detected."""
        if not callable(func):
            raise ValueError("Provided handler func is not callable")
        self._falling_handler = func
        self.pin.irq(handler=self._bounce_wrapper)

    def reset_handler(self):
        self.pin.irq(handler=None)

    def _handler_both(self, other, func):
        """When this and other are high, execute the both func."""
        if not callable(func):
            raise ValueError("Provided handler func is not callable")
        self._other = other
        self._both_handler = func
        self.pin.irq(handler=self._bounce_wrapper)


class DigitalInput(DigitalReader):
    """A class for handling reading of the digital input.

    The Digital Input jack can detect a HIGH signal when recieving voltage >
    0.8v and will be LOW when below.

    To use the handler method, you simply define whatever you want to happen
    when a button or the digital input is triggered, and then use
    ``x.handler(new_function)``. Do not include the brackets for the function,
    and replace the 'x' in the example with the name of your input, either
    ``b1``, ``b2``, or ``din``.

    Here is another example how you can write digital input handlers to react
    to a clock source and match its trigger duration.::

        @din.handler
        def gate_on():
            # Trigger outputs with a probability set by knobs.
            cv1.value(random() > k1.percent())
            cv2.value(random() > k2.percent())

        @din.handler_falling
        def gate_off():
            # Turn off all triggers on falling clock trigger to match clock.
            cv1.off()
            cv2.off()

    When writing a handler, try to keep the code as minimal as possible.
    Ideally handlers should be used to change state and allow your main loop
    to change behavior based on the altered state. See `tips <https://docs.micropython.org/en/latest/reference/isr_rules.html#tips-and-recommended-practices>`_
    from the MicroPython documentation for more details.

    :param pin:  The pin ID we read from
    :param debounce_delay:  The time in ms that we debounce the input
    """

    def __init__(self, pin, debounce_delay=0):
        super().__init__(pin, debounce_delay)

    def last_triggered(self):
        """Return the ticks_ms of the last trigger.

        If the button has not yet been pressed, the default return value is 0.
        """
        return self.last_rising_ms


class Button(DigitalReader):
    """A class for handling push button behavior.

    Button instances have a method ``last_pressed()``
    (similar to ``DigitalInput.last_triggered()``) which can be used by your
    script to help perform some action or behavior relative to when the button
    was last pressed (or input trigger received). For example, if you want to
    call a function to display a message that a button was pressed, you could
    add the following code to your main script loop::

        # Inside the main loop...
        if b1.last_pressed() > 0 and ticks_diff(ticks_ms(), b1.last_pressed()) < 2000:
            # Call this during the 2000 ms duration after button press.
            display_button_pressed()

    Note, if a button has not yet been pressed, the ``last_pressed()`` default
    return value is 0, so you may want to add the check `if b1.last_pressed() > 0`
    before you check the elapsed duration to ensure the button has been
    pressed. This is also useful when checking if the digital input has been
    triggered with the ``DigitalInput.last_triggered()`` method.

    :param pin:  The pin ID we read from
    :param debounce_delay:  The time in ms that we debounce the input
    """

    def __init__(self, pin, debounce_delay=200):
        super().__init__(pin, debounce_delay)

    def last_pressed(self):
        """Return the ticks_ms of the last button press

        If the button has not yet been pressed, the default return value is 0.
        """
        return self.last_rising_ms


class Output:
    """A class for sending digital or analogue voltage to an output jack.

    The outputs are capable of providing 0-10V, which can be achieved using
    the ``cvx.voltage()`` method.

    So that there is no chance of not having the full range, the chosen
    resistor values actually give you a range of about 0-10.5V, which is why
    calibration is important if you want to be able to output precise voltages.

    :param pin:  The GPIO pin we use as a PWM output
    :param min_voltage: The minimum allowed output voltage
    :param max_voltage: The maximum allowed output voltage
    :param gate_voltage:  The voltage we use for gate signals (see ``.on()``, ``.off()``)
    :param calibration_values:  Calibration data for this output
    """

    def __init__(
        self,
        pin,
        min_voltage=MIN_OUTPUT_VOLTAGE,
        max_voltage=MAX_OUTPUT_VOLTAGE,
        gate_voltage=GATE_VOLTAGE,
        calibration_values=OUTPUT_CALIBRATION_VALUES[0],
    ):
        self.pin = PWM(Pin(pin))
        self.pin.freq(PWM_FREQ)
        self.MIN_VOLTAGE = min_voltage
        self.MAX_VOLTAGE = max_voltage
        self.gate_voltage = gate_voltage

        self._calibration_values = calibration_values
        self._duty = 0
        self._gradients = []
        for index, value in enumerate(self._calibration_values[:-1]):
            self._gradients.append(self._calibration_values[index + 1] - value)
        self._gradients.append(self._gradients[-1])

    def _set_duty(self, cycle):
        cycle = int(cycle)
        self.pin.duty_u16(clamp(cycle, 0, MAX_UINT16))
        self._duty = cycle

    def voltage(self, voltage=None):
        """
        Set the output voltage to the provided value within the range of MIN_VOLTAGE to MAX_VOLTAGE

        By default this range is 0-10, but can be overridden via the configuration file

        :param voltage:  The desired volts to send to the output. If None the current voltage is returned
        """
        if voltage is None:
            return self._duty / MAX_UINT16 * self.MAX_VOLTAGE
        voltage = clamp(voltage, self.MIN_VOLTAGE, self.MAX_VOLTAGE)
        index = int(voltage // 1)
        self._set_duty(self._calibration_values[index] + (self._gradients[index] * (voltage % 1)))

    def on(self):
        """
        Set the voltage HIGH according to the gate voltage

        By default this is 5V, but can be overridden via the configuration file
        """
        self.voltage(self.gate_voltage)

    def off(self):
        """Set the voltage LOW at 0 volts."""
        self._set_duty(0)

    def toggle(self):
        """Invert the Output's current state."""
        if self._duty > 500:
            self.off()
        else:
            self.on()

    def value(self, value):
        """
        Sets the output to 0V or 5V based on a binary input, 0 or 1.

        :param value:  HIGH or LOW, according to the desired state.
        """
        if value:  # silently allow booleans too!
            self.on()
        else:
            self.off()


class Thermometer:
    """
    Wrapper for the temperature sensor connected to Pin 4

    Reports the module's current temperature in Celsius.

    If the module's temperature sensor is not working correctly, the temperature will always be reported as None
    """

    # Conversion factor for converting from the raw ADC reading to sensible units
    # See Raspberry Pi Pico datasheet for details
    TEMP_CONV_FACTOR = 3.3 / 65535

    def __init__(self):
        # The Raspberry Pi Pico 2's temperature sensor doesn't work with some older
        # uPython firmwares so do some basic exception handling
        try:
            self.pin = ADC(PIN_TEMPERATURE)
        except:
            self.pin = None

    def read_temperature(self):
        """
        Read the ADC and return the current temperature

        :return:  The current temperature in Celsius, or None if the hardware did not initialze properly
        """
        if self.pin:
            # See the Pico's datasheet for the details of this calculation
            return 27 - ((self.pin.read_u16() * self.TEMP_CONV_FACTOR) - 0.706) / 0.001721
        else:
            return None


class UsbConnection:
    """
    Checks the USB terminal is connected or not

    On the original Pico we can check Pin 24, but on the Pico 2 this does not work. In that case
    check the SIE_STATUS register and check bit 16
    """

    def __init__(self):
        if europi_config.PICO_MODEL == "pico2":
            self.pin = None
        else:
            self.pin = DigitalReader(PIN_USB_CONNECTED)

    def value(self):
        """Return 0 or 1, indicating if the USB connection is disconnected or connected"""
        if self.pin:
            return self.pin.value()
        else:
            # see https://forum.micropython.org/viewtopic.php?t=10814#p59545
            SIE_STATUS = 0x50110000 + 0x50
            BIT_CONNECTED = 1 << 16

            if mem32[SIE_STATUS] & BIT_CONNECTED:
                return 1
            else:
                return 0


# Define all the I/O using the appropriate class and with the pins used
din = DigitalInput(PIN_DIN)
ain = AnalogueInput(PIN_AIN)
k1 = Knob(PIN_K1)
k2 = Knob(PIN_K2)
b1 = Button(PIN_B1)
b2 = Button(PIN_B2)

# Output CVs
cv1 = Output(PIN_CV1, calibration_values=OUTPUT_CALIBRATION_VALUES[0])
cv2 = Output(PIN_CV2, calibration_values=OUTPUT_CALIBRATION_VALUES[1])
cv3 = Output(PIN_CV3, calibration_values=OUTPUT_CALIBRATION_VALUES[2])
cv4 = Output(PIN_CV4, calibration_values=OUTPUT_CALIBRATION_VALUES[3])
cv5 = Output(PIN_CV5, calibration_values=OUTPUT_CALIBRATION_VALUES[4])
cv6 = Output(PIN_CV6, calibration_values=OUTPUT_CALIBRATION_VALUES[5])
cvs = [cv1, cv2, cv3, cv4, cv5, cv6]

# Helper object for reading the onboard temperature sensor
thermometer = Thermometer()

# Helper object to detect if the USB cable is connected or not
usb_connected = UsbConnection()

# External I2C
external_i2c = I2C(
    europi_config.EXTERNAL_I2C_CHANNEL,
    sda=Pin(europi_config.EXTERNAL_I2C_SDA),
    scl=Pin(europi_config.EXTERNAL_I2C_SCL),
    freq=europi_config.EXTERNAL_I2C_FREQUENCY,
    timeout=europi_config.EXTERNAL_I2C_TIMEOUT,
)

# Set the desired clock speed according to the configuration
# By default this will overclock the CPU, but some users may not want to
# e.g. to lower power consumption on a very power-constrained system
freq(CPU_FREQS[europi_config.PICO_MODEL][europi_config.CPU_FREQ])



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface for realtime clock support.

EuroPi can have 0-1 realtime clocks at a time. The implementation is specified via experimental_config.
Underlying implementations to handle network and/or I2C hardware is contained with the experimental.clocks
namespace.

This module reads the desired implementation from experimental_config and instantiates a clock object
that can be used externally.
"""

import europi
from experimental.clocks.clock_source import ExternalClockSource
from experimental.experimental_config import RTC_DS1307, RTC_DS3231, RTC_NTP


class Month:
    """
    Container class for month names
    """

    JANUARY = 1
    FEBRUARY = 2
    MARCH = 3
    APRIL = 4
    MAY = 5
    JUNE = 6
    JULY = 7
    AUGUST = 8
    SEPTEMBER = 9
    OCTOBER = 10
    NOVEMBER = 11
    DECEMBER = 12

    NAME = {
        1: "January",
        2: "February",
        3: "March",
        4: "April",
        5: "May",
        6: "June",
        7: "July",
        8: "August",
        9: "September",
        10: "October",
        11: "November",
        12: "December",
    }


class Weekday:
    """
    Container class for weekday names

    ISO 8601 specifies the week starts on Monday (1) and ends on Sunday (7)
    """

    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7

    NAME = {
        0: "Unspecified",
        1: "Monday",
        2: "Tuesday",
        3: "Wednesday",
        4: "Thursday",
        5: "Friday",
        6: "Saturday",
        7: "Sunday",
    }


class Timezone:
    """
    Represents a timezone shift relative to UTC

    We allow minutes & hours, since there are 15-, 30-, and 45-minute timzeones in the world
    """

    def __init__(self, hours, minutes):
        """
        Create a time zone we can add to a datetime to get local time

        :param hours:  The number of hours ahead/behind we need to adjust [-24 to +24]
        :param minutes:  The number of minutes ahead/behind we need to adjust [-59 to +59]
        """
        if (hours < 0 and minutes > 0) or (hours > 0 and minutes < 0):
            raise ValueError("Timezone offset must be in a consistent direction")
        if abs(hours) > 24 or abs(minutes) > 59:
            raise ValueError("Invalid time zone adjustment")
        self.hours = hours
        self.minutes = minutes

    def __str__(self):
        """
        Get the offset as a string

        Result is of the format is {+|-}hh:mm
        """
        s = f"{abs(self.hours):02}:{abs(self.minutes):02}"
        if self.hours < 0:
            s = f"-{s}"
        else:
            s = f"+{s}"
        return s


class DateTime:
    """
    Represents a date and time

    :param year:  The current year (e.g. 2025)
    :param month:  The current month (e.g. Month.JANUARY)
    :param day:  The current day within the month (e.g. 17)
    :param hour:  The current hour on a 24-hour clock (0-23)
    :param minute:  The current minute within the hour (0-59)
    :param second:  The current second within the minute (0-59, optional)
    :param weekday:  The current day of the week (e.g. Weekday.MONDAY, optional)
    :param yearday:  The current day of the year (1-365, 1-366 if leap year, optional)
    """

    def __init__(self, year, month, day, hour, minute, second=None, weekday=None, yearday=None):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second
        self.weekday = weekday

    def __str__(self):
        """
        Get the current day and time as a string

        Result is of the format
            [Weekday ]YYYY/MM/DD hh:mm[:ss]
        """
        s = f"{self.year:04}/{self.month:02}/{self.day:02} {self.hour:02}:{self.minute:02}"
        if self.second is not None:
            s = f"{s}:{self.second:02}"
        if self.weekday is not None:
            s = f"{Weekday.NAME[self.weekday]} {s}"
        return s

    def __add__(self, tz):
        """
        Add a timezone offset to the current time, returning the result

        :param tz:  The timezone we're adding to this Datetime
        """
        t = DateTime(
            self.year, self.month, self.day, self.hour, self.minute, self.second, self.weekday
        )

        # shortcut if there is no offset
        if tz.hours == 0 and tz.minutes == 0:
            return t

        # temporarily shift the month and day to be zero-indexed to make subsequent math easier
        t.month -= 1
        t.day -= 1
        if t.weekday is not None:
            t.weekday -= 1

        # add the offset; this can be positive or negative
        t.minute += tz.minutes
        t.hour += tz.hours

        # cascade through the units, borrowing/carrying-over as needed
        if t.minute < 0:
            t.minute += 60
            t.hour -= 1
        elif t.minute >= 60:
            t.minute -= 60
            t.hour += 1

        if t.hour < 0:
            t.hour += 24
            t.day -= 1
            if t.weekday is not None:
                t.weekday = (t.weekday - 1) % 7
        elif t.hour >= 24:
            t.hour -= 24
            t.day += 1
            if t.weekday is not None:
                t.weekday = (t.weekday + 1) % 7

        days_in_month = DateTime.calculate_days_in_month(t.month + 1, t.year)
        days_in_prev_month = DateTime.calculate_days_in_month((t.month + 1) % 12 + 1, t.year)
        if t.day < 0:
            t.day = days_in_prev_month - 1  # last day of the month, zero-indexed
            t.month -= 1
        elif t.day >= days_in_month:
            t.day = 1
            t.month += 1

        if t.month < 0:
            t.month += 12
            t.year -= 1
        elif t.month >= 12:
            t.month -= 12
            t.year += 1

        # shift the month, day and weekday back to be 1-indexed
        t.month += 1
        t.day += 1
        if t.weekday is not None:
            t.weekday += 1

        return t

    @staticmethod
    def calculate_days_in_month(month, year):
        month_lengths = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        if DateTime.calculate_is_leap_year(year) and month == Month.FEBRUARY:
            return 29
        else:
            return month_lengths[month - 1]

    @staticmethod
    def calculate_is_leap_year(year):
        # a year is a leap year if it is divisible by 4
        return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)

    @property
    def is_leap_year(self):
        return DateTime.calculate_is_leap_year(self.year)

    @property
    def days_in_month(self):
        return DateTime.calculate_days_in_month(self.month, self.year)

    @property
    def days_in_year(self):
        if self.is_leap_year:
            return 366
        else:
            return 365

    @property
    def tuple(self):
        return (
            self.year,
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            self.weekday,
            0,
        )

    def __eq__(self, other):
        # fmt: off
        return (
            self.year == other.year and
            self.month == other.month and
            self.day == other.day and
            self.hour == other.hour and
            self.minute == other.minute and
            (
                (self.second is None or other.hour is None ) or
                (self.second == other.second)
            )
        )
        # fmt: on

    def __lt__(self, other):
        if self.year == other.year:
            if self.month == other.month:
                if self.hour == other.hour:
                    if self.minute == other.minute:
                        if self.second is None or other.second is None:
                            return False
                        return self.second < other.second
                    return self.minute < other.minute
                return self.hour < other.hour
            return self.month < other.month
        return self.year < other.year

    def __gt__(self, other):
        if self.year == other.year:
            if self.month == other.month:
                if self.hour == other.hour:
                    if self.minute == other.minute:
                        if self.second is None or other.second is None:
                            return False
                        return self.second > other.second
                    return self.minute > other.minute
                return self.hour > other.hour
            return self.month > other.month
        return self.year > other.year

    def __le__(self, other):
        if self.year == other.year:
            if self.month == other.month:
                if self.hour == other.hour:
                    if self.minute == other.minute:
                        if self.second is None or other.second is None:
                            return False
                        return self.second <= other.second
                    return self.minute <= other.minute
                return self.hour <= other.hour
            return self.month <= other.month
        return self.year <= other.year

    def __ge__(self, other):
        if self.year == other.year:
            if self.month == other.month:
                if self.hour == other.hour:
                    if self.minute == other.minute:
                        if self.second is None or other.second is None:
                            return False
                        return self.second >= other.second
                    return self.minute >= other.minute
                return self.hour >= other.hour
            return self.month >= other.month
        return self.year >= other.year


class RealtimeClock:
    """
    A continually-running clock that provides the day & date.

    This class wraps around an external clock source, e.g. an I2C-compatible RTC
    module or a network connection to an NTP server

    :param source:  An ExternalClockSource implementation we read the time from
    """

    def __init__(self, source):
        self.source = source

    def utcnow(self):
        """
        Get the current UTC time.

        :return: A DateTime object representing the current UTC time
        """

        # get the raw tuple from the clock, append Nones so we have all 7 fields
        t = list(self.source.datetime())
        if len(t) < ExternalClockSource.SECOND + 1:
            t.append(None)
        if len(t) < ExternalClockSource.WEEKDAY + 1:
            t.append(None)

        return DateTime(
            t[ExternalClockSource.YEAR],
            t[ExternalClockSource.MONTH],
            t[ExternalClockSource.DAY],
            t[ExternalClockSource.HOUR],
            t[ExternalClockSource.MINUTE],
            t[ExternalClockSource.SECOND],
            t[ExternalClockSource.WEEKDAY],
            t[ExternalClockSource.YEARDAY],
        )

    def localnow(self):
        """
        Get the current local time

        See experimental_config for instructions on how to configure the local timezone

        :return: a DateTime object representing the current local time
        """
        return self.utcnow() + local_timezone


# fmt: off
if europi.experimental_config.RTC_IMPLEMENTATION == RTC_DS1307:
    from experimental.clocks.ds1307 import DS1307
    source = DS1307(europi.external_i2c)
elif europi.experimental_config.RTC_IMPLEMENTATION == RTC_DS3231:
    from experimental.clocks.ds3231 import DS3231
    source = DS3231(europi.external_i2c)
elif europi.experimental_config.RTC_IMPLEMENTATION == RTC_NTP:
    from experimental.clocks.ntp import NtpClock
    source = NtpClock()
else:
    from experimental.clocks.null_clock import NullClock
    source = NullClock()
# fmt: on

local_timezone = Timezone(
    europi.experimental_config.UTC_OFFSET_HOURS, europi.experimental_config.UTC_OFFSET_MINUTES
)

clock = RealtimeClock(source)



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module provides reusable UI components.
"""

from europi import CHAR_HEIGHT, CHAR_WIDTH, OLED_HEIGHT, b1, k1, oled


class Menu:
    """A class representing a menu displayable on the EuroPi screen. The user can select a menu item using the
    configured knob and select it with the configured button(s).

    :param items: a list of the menu items
    :param select_func: the function to call when a menu item is chosen. The function will be called with single argument, the selected item.
    :param select_knob: the knob used to select the menu item, defaults to k1
    :param choice_buttons: a List of Buttons that can be pressed to choose the selected item. Defaults to b1
    """

    def __init__(self, items, select_func, select_knob=k1, choice_buttons=None):
        self.items = ["----- MENU -----"] + items
        self.select_func = select_func
        self.select_knob = select_knob
        choice_buttons = choice_buttons or [b1]

        # init handlers
        def select():
            self.select_func(self.items[self.selected + 1])

        for b in choice_buttons:
            b.handler_falling(select)

    @property
    def selected(self):
        """The currently selected menu item."""
        return self.select_knob.read_position(steps=len(self.items) - 1)

    def _inverted_text(self, s, x, y):
        """displays the given text with an inverted background"""
        oled.fill_rect(x, y - 1, CHAR_WIDTH * len(s), CHAR_HEIGHT + 2, 1)
        oled.text(s, x, y, 0)

    def draw_menu(self):
        """This function should be called by your script's main loop in order to display and refresh the menu."""
        current = self.selected
        oled.fill(0)
        line_height = CHAR_HEIGHT + 2

        for line_number, item in enumerate(self.items):
            y_position = ((line_number - current) * line_height) + 1
            if (
                y_position <= (OLED_HEIGHT - line_height) and y_position > 0
            ):  # Only draw lines which can be fully displayed
                if line_number == (current + 1):
                    self._inverted_text(f"{item}", 2, y_position)
                else:
                    oled.text(f"{item}", 2, y_position, 1)

        oled.show()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for searching, splitting, and inserting into a sorted list

Copied from https://github.com/python/cpython/blob/3.12/Lib/bisect.py with some modifications
noted in-line with `# EDIT`
"""


def insort_right(a, x, lo=0, hi=None, *, key=None):
    """Insert item x in list a, and keep it sorted assuming a is sorted.

    If x is already in a, insert it to the right of the rightmost x.

    :param a: The array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """
    if key is None:
        lo = bisect_right(a, x, lo, hi)
    else:
        lo = bisect_right(a, key(x), lo, hi, key=key)
    a.insert(lo, x)


def bisect_right(a, x, lo=0, hi=None, *, key=None):
    """Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e <= x, and all e in
    a[i:] have e > x.  So if x already appears in the list, a.insert(i, x) will
    insert just after the rightmost x already there.

    :param a: The sorted array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """

    if lo < 0:
        # EDIT
        # raise ValueError('lo must be non-negative')
        lo = 0
    if hi is None:
        hi = len(a)
    # Note, the comparison uses "<" to match the
    # __lt__() logic in list.sort() and in heapq.
    if key is None:
        while lo < hi:
            mid = (lo + hi) // 2
            if x < a[mid]:
                hi = mid
            else:
                lo = mid + 1
    else:
        while lo < hi:
            mid = (lo + hi) // 2
            if x < key(a[mid]):
                hi = mid
            else:
                lo = mid + 1
    return lo


def insort_left(a, x, lo=0, hi=None, *, key=None):
    """Insert item x in list a, and keep it sorted assuming a is sorted.

    If x is already in a, insert it to the left of the leftmost x.

    :param a: The sorted array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """

    if key is None:
        lo = bisect_left(a, x, lo, hi)
    else:
        lo = bisect_left(a, key(x), lo, hi, key=key)
    a.insert(lo, x)


def bisect_left(a, x, lo=0, hi=None, *, key=None):
    """Return the index where to insert item x in list a, assuming a is sorted.

    The return value i is such that all e in a[:i] have e < x, and all e in
    a[i:] have e >= x.  So if x already appears in the list, a.insert(i, x) will
    insert just before the leftmost x already there.

    :param a: The sorted array we're inserting into
    :param x: The value we're inserting into the array
    :param lo: the lower bound of the array slice to consider
    :param hi: the upper bound of the array slice to consider
    :param key: an optional key function for sorting the data
    """

    if lo < 0:
        # EDIT
        # raise ValueError('lo must be non-negative')
        lo = 0
    if hi is None:
        hi = len(a)
    # Note, the comparison uses "<" to match the
    # __lt__() logic in list.sort() and in heapq.
    if key is None:
        while lo < hi:
            mid = (lo + hi) // 2
            if a[mid] < x:
                lo = mid + 1
            else:
                hi = mid
    else:
        while lo < hi:
            mid = (lo + hi) // 2
            if key(a[mid]) < x:
                lo = mid + 1
            else:
                hi = mid
    return lo


# Create aliases
bisect = bisect_right
insort = insort_right



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Open Sound Control over UDP implementation.

See
- https://opensoundcontrol.stanford.edu/
- https://hexler.net/touchosc/manual/introduction
"""

from europi_log import *
import socket
import struct


def align_next_word(n):
    """
    Return the index of the next word-alined byte

    We assume 4-byte/32-bit words. If we're already word-aligned,
    we increment to the next one

    :param n:  The current index in a byte array
    """
    return n + (4 - (n % 4)) % 4


class OpenSoundPacket:
    """
    A container object for the Open Sound Control packet(s) we receive.

    Contains the address of the message and the data

    :property address:  The address string of the packet
    :property values:  The values included in the packet

    :param data:  The raw byte data read from the UDP socket. The byte data consists
        of the following data:
        #. leading '/' character

        #. slash-separated address (e.g. foo/bar)

        #. null terminator (0x00 byte)

        #. ',' character

        #. type arguments

            #. "f" character for 32-bit float

            #. "i" character for 32-bit signed integer

            #. "s" character for string (null-terminated)

            #. "b" character for blob (32-bit int -> n, followed by n bytes of data)

            #. assorted non-standard types

        #. null terminator(s)

        #. payload bytes (lengths are type dependent)

        Every argument starts on an 4-aligned byte, so there are
        filler nulls to pad strings out to a multiple of 32 bits
    """

    def __init__(self, data: bytes):
        address_end = data.index(b"\0", 1)
        self._address = data[0:address_end].decode("utf-8")
        if self._address.endswith("/"):
            self._address = self._address.rstrip("/")

        types = []
        self._values = []
        type_start = data.index(b",", address_end)
        data_start = data.index(b"\0", type_start)
        data_start = align_next_word(data_start)
        i = type_start + 1
        d = data_start
        while data[i] != 0x00:
            t = chr(data[i])
            if t == "i":
                types.append(int)
                n = (data[d] << 24) | (data[d + 1] << 16) | (data[d + 2] << 8) | data[d + 1]
                self.values.append(n)
                d += 4
            elif t == "f":
                types.append(float)
                n = struct.unpack(">f", data[d : d + 4])[0]
                self.values.append(n)
                d += 4
            elif t == "s" or t == "S":  # include the alternate "S" here
                types.append(str)
                s = ""
                string_end = data.index(b"\0", d)
                for c in range(d, string_end):
                    s += chr(data[c])
                self.values.append(s)
                d = string_end
                d = align_next_word(d)
            elif t == "b":
                # blob; int32 -> n, followed by n bytes
                types.append(bytearray)
                n = (data[d] << 24) | (data[d + 1] << 16) | (data[d + 2] << 8) | data[d + 1]
                d += 4
                b = []
                for i in range(n):
                    b.append(data[d + i])
                d += n
                d = align_next_word(d)
                self.values.append(bytearray(b))
            elif t == "T" or t == "F":
                # zero-byte boolean; skip
                pass
            elif t == "t":
                # 8-byte timestamp; skip
                d += 8
            elif t == "h":
                # 64-bit signed integer
                # treat as a normal int
                types.append(int)
                n = (
                    (data[d] << 56)
                    | (data[d + 1] << 48)
                    | (data[d + 2] << 40)
                    | (data[d + 3] << 32)
                    | (data[d + 4] << 24)
                    | (data[d + 5] << 16)
                    | (data[d + 6] << 8)
                    | data[d + 7]
                )
                self.values.append(n)
                d += 8
            elif t == "c":
                # a single character; treat as a string
                types.append(str)
                self.values.append(
                    data[d + 3].decode()  # data is in the 4th byte; padded with leading zeros
                )
                d += 4
            elif t == "m":
                # 4-byte midi; skip
                d += 4
            elif t == "N":
                # nil; skip
                pass
            elif t == "I":
                # infinity; skip
                pass
            else:
                log_warning(f"Unsupported type {t}", "osc")

            i += 1

    @property
    def values(self) -> list[int | float | str | bytearray]:
        """
        The array of values included in this packet

        We support OSC 1.0 + a subset of 1.1 types
        """
        return self._values

    @property
    def address(self) -> str:
        """This packet's address"""
        return self._address


class OpenSoundServer:
    """
    The OSC server.

    This object owns the socket and is responsible for sending & receiving
    messages

    To use, instantiate the server and assign a callback. Then call
    receive_data() in the main loop:

    .. code-block:: python

        import europi

        def main():
            srv = OpenSoundServer(9000)

            @srv.data_handler
            def process_data(connection=None, data=None):
                # process the data, send anything back over the connection if needed
                ...

            while True:
                srv.receive_data()

    :param recv_port:  The UDP port we accept messages on. TouchOSC uses port 9000 by default,
        so we use that here for convenience
    :param send_port:  The UDP port we send outgoing messages on.
    :param send_addr:  The IP address of the host we send outgoing messages to
    """

    def __init__(self, recv_port=9000, send_port=9001, send_addr="192.168.4.100"):
        log_info(f"Listening for OSC packets on port {recv_port}", "osc")
        addr = socket.getaddrinfo("0.0.0.0", recv_port)[0][-1]
        self.recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.recv_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.recv_socket.settimeout(0)
        self.recv_socket.bind(addr)

        log_info(f"Sending OSC packets to {send_addr}:{send_port}", "osc")
        addr = socket.getaddrinfo(send_addr, send_port)[0][-1]
        self.send_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.send_port = send_port
        self.send_addr = send_addr

        self.recv_callback = self.default_callback

    def default_callback(self, connection=None, data: OpenSoundPacket = None):
        """
        Default callback function when we receive data

        Does nothing
        """
        pass

    def data_handler(self, func):
        """
        Decorator for the function to UDP packet reception

        The provided function must accept the following keyword arguments:
        - connection: socket  A socket connection to the client
        - data: str  The request the client sent

        :param func:  The function to handle the request.
        """

        def wrapper(*args, **kwargs):
            func(*args, **kwargs)

        self.recv_callback = wrapper
        return wrapper

    def receive_data(self):
        """Check if we have any new data to process, invoke data_handler as needed"""
        while True:
            try:
                (data, connection) = self.recv_socket.recvfrom(1024)
                packet = OpenSoundPacket(data)
                self.recv_callback(connection=connection, data=packet)
            except ValueError as err:
                log_warning(f"Failed to process packet: {err}", "osc")
                break
            except OSError as err:
                break
            except Exception as err:
                # log_warning(f"Failed to process packet. Malformed? {err}", "osc")
                # s = ""
                # for byte in data:
                #     s += f"{byte:02x} "
                # log_debug(f"Raw packet: {s}")
                break

    def send_data(self, address, *args):
        """
        Transmit a packet

        :param address:  The OSC address to send to
        :param args:  The values to encode in the packet. Allowed types are int, float, bool, str, and bytearray. Bools are converted to 0/1 integers
        """

        def pad_length(arr):
            # pad with nulls until we get to the next multiple of 4
            for i in range((4 - (len(arr) % 4)) % 4):
                arr.append(0)

        data = []
        for ch in address:
            data.append(ord(ch))
        data.append(0)
        pad_length(data)

        # comma denoting start of types
        data.append(ord(","))

        # types
        for arg in args:
            if type(arg) is int or type(arg) is bool:
                data.append(ord("i"))
            elif type(arg) is float:
                data.append(ord("f"))
            elif type(arg) is str:
                data.append(ord("s"))
            elif type(arg) is bytearray:
                data.append(ord("b"))
        data.append(0)
        pad_length(data)

        # values
        for arg in args:
            if type(arg) is int:
                # big-endian, 32-bit int
                data.append((arg >> 24) & 0xFF)
                data.append((arg >> 16) & 0xFF)
                data.append((arg >> 8) & 0xFF)
                data.append(arg & 0xFF)
            elif type(arg) is bool:
                # treat boolean as a 0/1 integer for convenience
                if arg:
                    data.append(0)
                    data.append(0)
                    data.append(0)
                    data.append(1)
                else:
                    data.append(0)
                    data.append(0)
                    data.append(0)
                    data.append(0)
            elif type(arg) is float:
                bytes = struct.pack(">f", arg)
                for b in bytes:
                    data.append(b)
            elif type(arg) is str:
                for ch in arg:
                    data.append(ord(ch))
                data.append(0)
                pad_length(data)
            elif type(arg) is bytearray:
                n = len(arg)
                data.append((n >> 24) & 0xFF)
                data.append((n >> 16) & 0xFF)
                data.append((n >> 8) & 0xFF)
                data.append(n & 0xFF)
                for b in arg:
                    data.append(b)
                pad_length(data)

        try:
            data = bytearray(data)
            self.send_socket.sendto(data, (self.send_addr, self.send_port))
        except OSError:
            pass
        except Exception as err:
            log_warning(f"Failed to send OSC data: {err}", "osc")



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from time import sleep
from europi import oled, b1, b2, k2, ain, cvs, usb_connected, turn_off_all_cvs
from europi_script import EuroPiScript
from os import stat, mkdir
from experimental.math_extras import mean


class CalibrationValues:
    """Wrapper class for the input & output calibration values used for analogue inputs & outputs

    In low-accuracy mode, input_calibration_values is a length 2 array with the raw samples taken at 0V and 10V

    In high-accuracy mode, input_calibration_values is a length 11 array with the raw samples taken at 0-10V, in 1V
    increments.

    In either mode, output_calibration_values is a length 11 array with the raw samples taken at 0-10V, in 1V
    increments.

    :param mode:  The calibration mode, one of MODE_LOW_10, MODE_LOW_5, or MODE_HIGH
    """

    MODE_UNKNOWN = "unk"
    MODE_LOW_10V = "low10"
    MODE_LOW_5V = "low5"
    MODE_HIGH = "high"

    mode = "unk"

    input_calibration_values = []
    output_calibration_values = []

    def __init__(self, mode):
        self.mode = mode

    def save(self):
        """Save the calibration readings to /lib/calibration_values.py

        Note: this will overwrite all previous calibrations
        """
        with open(f"lib/calibration_values.py", "w") as file:
            values = ", ".join(map(str, self.input_calibration_values))
            file.write(f"INPUT_CALIBRATION_VALUES=[{values}]\n")

            values = ", ".join(map(str, self.output_calibration_values))
            file.write(f"OUTPUT_CALIBRATION_VALUES=[{values}]\n")

            file.write(f"CALIBRATION_MODE = '{self.mode}'\n")


class Calibrate(EuroPiScript):
    """
    A script to interactively calibrate the module

    General flow:
    #. Sanity check (rack power, necessary file structure exists)

    #. Input calibration. One of:

        * low-accuracy 10V in

        * low-accuracy 5V in

        * high-accuracy 0-10V in

    #. Output calibration

    #. Save calibration

    #. Idle for reboot

    Output calibration only applies to CV1; all other outputs will
    use the same calibration values.
    """

    # The initial state; prompt the user to select their calibration mode using K2 + B2
    STATE_STARTUP = 0

    # Asking the user to select their calibration mode
    STATE_MODE_SELECT = 1

    # Initial states for each input calibration mode:
    # - low-accuracy w/ 10V input
    # - low-accuracy w/ 5V input
    # - high-accuracy w/ variable 0-10V input
    STATE_START_LOW_10 = 2
    STATE_START_LOW_5 = 3
    STATE_START_HIGH = 4

    # Start output calibration
    STATE_START_OUTPUT = 5

    # Flags indicating we're waiting for the user to press B1
    STATE_WAITING_B1 = -1
    STATE_B1_PRESSED = -2

    # As above, but for B2
    STATE_WAITING_B2 = -3
    STATE_B2_PRESSED = -4

    # The current state of the program
    # The state progresses pretty linearly with minimal branching, so mostly
    # this is used for debugging.
    # It _is_ necessary for button press detection
    state = 0

    def __init__(self):
        super().__init__()

    @classmethod
    def display_name(cls):
        """Push this script to the end of the menu."""
        return "~Calibrate"

    def text_wait(self, text, duration):
        """
        Display text on the screen and block for the specified duration

        :param text:  The text to display
        :param duration:  The duration to wait in seconds
        """
        oled.centre_text(text)
        sleep(duration)

    def read_sample(self):
        """
        Read from the raw ain pin and return the average across several readings

        :return:  The average across several distinct readings from the pin
        """
        N_READINGS = 512
        readings = []
        for i in range(N_READINGS):
            readings.append(ain.pin.read_u16())

        # discard the lowest & highest 1/4 of the readings as outliers
        readings.sort()
        readings = readings[N_READINGS // 4 : 3 * N_READINGS // 4]
        return round(mean(readings))

    def wait_for_voltage(self, voltage):
        """
        Wait for the user to connect the desired voltage to ain & press b1

        :param voltage:  The voltage to instruct the user to connect. Used for display only

        :return:  The average samples read from ain (see read_sample)
        """
        if voltage == 0:
            oled.centre_text("Unplug all\n\nDone: Button 1")
        else:
            oled.centre_text(f"Plug in {voltage:0.1f}V\n\nDone: Button 1")
        self.wait_for_b1()
        oled.centre_text(f"Calibrating...\nAIN @ {voltage:0.1f}V")
        sleep(1)
        return self.read_sample()

    def wait_for_b1(self, wait_fn=None):
        """
        Wait for the user to press B1, returning to the original state when they have

        :param wait_fn:  A function to execute while waiting (e.g. refresh the UI)
        """
        if wait_fn is None:
            wait_fn = lambda: sleep(0.01)

        prev_state = self.state
        self.state = self.STATE_WAITING_B1
        while self.state == self.STATE_WAITING_B1:
            wait_fn()
        self.state = prev_state

    def wait_for_b2(self, wait_fn=None):
        """
        Wait for the user to press B2, returning to the original state when they have

        :param wait_fn:  A function to execute while waiting (e.g. refresh the UI)
        """
        if wait_fn is None:
            wait_fn = lambda: sleep(0.01)

        prev_state = self.state
        self.state = self.STATE_WAITING_B2
        while self.state == self.STATE_WAITING_B2:
            wait_fn()
        self.state = prev_state

    def check_directory(self):
        """Check if /lib exists. If it does not, create it"""
        try:
            stat("/lib")
        except OSError:
            mkdir("/lib")

    def check_rack_power(self):
        if usb_connected.value() == 1:
            oled.centre_text("Make sure rack\npower is on\nDone: Button 1")
            self.wait_for_b1()

    def input_calibration_low10(self):
        """
        Low-accuracy, 10V input calibration

        Prompt the user for 0 and 10V inputs only

        :return:  The sample readings for 0 and 10V
        """
        self.state = self.STATE_START_LOW_10
        readings = [
            self.wait_for_voltage(0),
            self.wait_for_voltage(10),
        ]
        return readings

    def input_calibration_low5(self):
        """
        Low-accuracy, 5V input calibration

        Prompt the user for 0 and 5V inputs only. The 5V reading is extrapolated to 10V.

        :return:  The sample readings for 0 and 10V
        """
        self.state = self.STATE_START_LOW_5
        readings = [
            self.wait_for_voltage(0),
            self.wait_for_voltage(5),
        ]

        # Extrapolate from 5V input to 10V, assuming a linear progression
        samples_per_volt = round((readings[-1] - readings[0]) / 5)
        readings[-1] = samples_per_volt * 10 + readings[0]

        return readings

    def input_calibration_high(self):
        """
        High accuracy input calibration

        User is prompted to connect 0, 1, 2, ..., 9, 10V

        :return:  The sample readings for 0 to 10V (inclusive)
        """
        self.state = self.STATE_START_HIGH
        readings = []
        for i in range(11):
            readings.append(self.wait_for_voltage(i))
        return readings

    def calibrate_output(self, cv_n, calibration_values, input_readings):
        """
        Send volts from CVx to AIN, adjusting the duty cycle so the output is correct.

        :param cv_n:  A value 0 <= cv_n < len(cvs) indicating which CV output we're calibrating
        :param calibration_values:  The array of calibration values we append our results to
        :param input_readings:  The duty cycles of AIN corresponding to 0, 1, 2, ..., 9, 10 volts
        """
        oled.centre_text(f"Plug CV{cv_n+1} into\nanalogue in\nDone: Button 1")
        self.wait_for_b1()

        # Initial calibration in steps of 50, rapidly to get close to the goal
        COARSE_STEP = 50

        # Intermediate calibration in steps of 5
        INTERMEDIATE_STEP = 5

        # Final fine calibration in steps of 1
        FINE_STEP = 1

        # always 0 duty for 0V out
        calibration_values.output_calibration_values.append([0])

        # Output 1-10V on each CV output & read the result on AIN
        # Adjust the duty cycle of the CV output until the input is within an acceptable range
        duty = 0
        cvs[cv_n].pin.duty_u16(duty)
        for volts, expected_reading in enumerate(input_readings[1:]):
            oled.centre_text(f"Calibrating...\n CV{cv_n+1} @ {volts+1}V\n1/3")
            sleep(1)
            duty = self.coarse_output_calibration(cvs[cv_n], expected_reading, duty, COARSE_STEP)
            oled.centre_text(f"Calibrating...\n CV{cv_n+1} @ {volts+1}V\n2/3")
            duty = self.fine_output_calibration(
                cvs[cv_n], expected_reading, duty, INTERMEDIATE_STEP, COARSE_STEP
            )
            oled.centre_text(f"Calibrating...\n CV{cv_n+1} @ {volts+1}V\n3/3")
            duty = self.fine_output_calibration(
                cvs[cv_n], expected_reading, duty, FINE_STEP, INTERMEDIATE_STEP
            )

            calibration_values.output_calibration_values[-1].append(duty)

        cvs[cv_n].off()

    def coarse_output_calibration(self, cv, goal_duty, start_duty, step_size):
        """
        Perform a fast, coarse calibration to bring the CV output's duty cycle somewhere close

        This will exit if either the calibration is within +/- the step_size OR if the measured duty
        cycle is higher than the goal (i.e. we've over-shot the goal)

        :param cv:  The CV output pin we're adjusting
        :param goal_duty:  The AIN duty cycle we're expecting to read
        :param start_duty:  The CVx duty cycle we're applying to the output initially
        :param step_size:  The amount by which we adjust the duty cycle up to reach the goal

        :return: The adjusted output duty cycle
        """
        read_duty = self.read_sample()
        duty = start_duty
        while abs(read_duty - goal_duty) > step_size and read_duty < goal_duty:
            duty += step_size
            cv.pin.duty_u16(duty)
            read_duty = self.read_sample()

        return duty

    def fine_output_calibration(self, cv, goal_duty, start_duty, step_size, prev_step_size):
        """
        Perform a slower, fine calibration to bring the CV output's duty cycle towards the goal

        This exits if the measured duty cycle is within +/- 2*step_size OR if we make
        2*prev_step_size adjustments

        :param cv:  The CV output pin we're adjusting
        :param goal_duty:  The AIN duty cycle we're expecting to read
        :param start_duty:  The CVx duty cycle we're applying to the output initially
        :param step_size:  The amount by which we adjust the duty cycle up/down to reach the goal
        :param prev_step_size:  The previous iteration's step size, used to limit how many adjustments we make

        :return: The adjusted output duty cycle
        """
        MAX_COUNT = 2 * prev_step_size
        count = 0
        read_duty = self.read_sample()
        duty = start_duty
        while abs(read_duty - goal_duty) >= 2 * step_size and count < MAX_COUNT:
            count += 1
            if read_duty < goal_duty:
                duty += step_size
            elif read_duty > goal_duty:
                duty -= step_size
            cv.pin.duty_u16(duty)
            sleep(0.1)
            read_duty = self.read_sample()

        return duty

    def main(self):
        # Button handlers
        @b1.handler
        def on_b1_press():
            if self.state == self.STATE_WAITING_B1:
                self.state = self.STATE_B1_PRESSED

        @b2.handler
        def on_b2_press():
            if self.state == self.STATE_WAITING_B2:
                self.state = self.STATE_B2_PRESSED

        # The available calibration modes (selected by k2)
        CALIBRATION_MODES = [
            "Low (10V in)",
            "Low (5V in)",
            "High (0-10V in)",
        ]

        self.text_wait("Calibration\nMode", 3)

        ######################################################################################################

        # Initial hardware & software checks
        self.check_directory()
        self.check_rack_power()

        ######################################################################################################

        # Calibration start
        self.state = self.STATE_MODE_SELECT
        refresh_ui = lambda: oled.centre_text(
            f"Choose mode (k2)\n{k2.choice(CALIBRATION_MODES)}\nDone: Button 2"
        )
        refresh_ui()
        self.wait_for_b2(refresh_ui)

        ######################################################################################################

        # Peform the requested input calibration
        choice = k2.choice(CALIBRATION_MODES)
        if choice == CALIBRATION_MODES[0]:
            calibration_values = CalibrationValues(CalibrationValues.MODE_LOW_10V)
            calibration_values.input_calibration_values = self.input_calibration_low10()
        elif choice == CALIBRATION_MODES[1]:
            calibration_values = CalibrationValues(CalibrationValues.MODE_LOW_5V)
            calibration_values.input_calibration_values = self.input_calibration_low5()
        else:
            calibration_values = CalibrationValues(CalibrationValues.MODE_HIGH)
            calibration_values.input_calibration_values = self.input_calibration_high()

        # make a local copy of the analogue-in readings, extrapolated if necessary,
        # that we can use to perform the output calibration
        if calibration_values.mode == CalibrationValues.MODE_HIGH:
            readings_in = list(calibration_values.input_calibration_values)
        else:
            # expand the raw calibration values if we were in a low-accuracy mode such that
            # we have an expected reading for every volt from 0-10
            readings_in = [calibration_values.input_calibration_values[0]]
            m = (
                calibration_values.input_calibration_values[1]
                - calibration_values.input_calibration_values[0]
            ) / 10
            c = calibration_values.input_calibration_values[0]
            for x in range(1, 10):
                readings_in.append(round((m * x) + c))
            readings_in.append(calibration_values.input_calibration_values[-1])

        ######################################################################################################

        # Output calibration
        self.state = self.STATE_START_OUTPUT
        calibration_values.output_calibration_values = []
        turn_off_all_cvs()
        for i in range(len(cvs)):
            self.calibrate_output(i, calibration_values, readings_in)

        ######################################################################################################

        # Save the result
        oled.centre_text("Saving\ncalibration...")
        calibration_values.save()

        # Prompt the user to reboot to apply the new calibration
        # Spin here forever if necessary
        while True:
            self.text_wait("Calibration\ncomplete", 3)
            self.text_wait("Reboot\nto apply\nnew calibration", 3)


if __name__ == "__main__":
    Calibrate().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import configuration
from configuration import ConfigFile, ConfigSpec


# sub-key constants for CPU_FREQS dict (see below)
# the Europi default is to overclock, so to avoid confusion about the default
# not being "default" just use a different word
# fmt: off
DEFAULT_FREQ = "normal"
OVERCLOCKED_FREQ = "overclocked"
UNDERCLOCKED_FREQ = "underclocked"
# fmt: on

# Supported Pico model types
MODEL_PICO = "pico"
MODEL_PICO_H = "pico h"
MODEL_PICO_W = "pico w"
MODEL_PICO_2 = "pico 2"
MODEL_PICO_2W = "pico 2w"

# Default & overclocked CPU frequencies for supported boards
# Key: board type (corresponds to EUROPI_MODEL setting)
# Sub-key: "default" or "overclocked" or "underclocked"
# fmt: off
CPU_FREQS = {
    MODEL_PICO: {
        DEFAULT_FREQ: 125_000_000,      # Pico default frequency is 125MHz
        OVERCLOCKED_FREQ: 250_000_000,  # Overclocked frequency is 250MHz
        UNDERCLOCKED_FREQ: 75_000_000   # Underclock to 75MHz
    },
    MODEL_PICO_2: {
        DEFAULT_FREQ: 150_000_000,      # Pico 2 default frequency is 150MHz
        OVERCLOCKED_FREQ: 300_000_000,  # Overclocked frequency is 300MHz
        UNDERCLOCKED_FREQ: 75_000_000,  # Underclock to 75MHz
    },
    MODEL_PICO_2W: {
        DEFAULT_FREQ: 150_000_000,      # Pico 2 W default frequency is 150MHz
        OVERCLOCKED_FREQ: 250_000_000,  # Overclocked frequency is 250MHz; 300MHz breaks wifi. See https://github.com/micropython/micropython/issues/16799
        UNDERCLOCKED_FREQ: 75_000_000,  # Underclock to 75MHz
    },
    MODEL_PICO_H: {
        DEFAULT_FREQ: 125_000_000,      # Pico H default frequency is 125MHz
        OVERCLOCKED_FREQ: 250_000_000,  # Overclocked frequency is 250MHz
        UNDERCLOCKED_FREQ: 75_000_000,  # Underclock to 75MHz
    },
    MODEL_PICO_W: {
        DEFAULT_FREQ: 125_000_000,      # Pico W default frequency is 125MHz
        OVERCLOCKED_FREQ: 250_000_000,  # Overclocked frequency is 250MHz
        UNDERCLOCKED_FREQ: 75_000_000,  # Underclock to 75MHz
    }
}
# fmt: on


class EuroPiConfig:
    """This class provides EuroPi's global config points.

    To override the default values, create /config/EuroPiConfig.json on the Raspberry Pi Pico
    and populate it with a JSON object.  See CONFIGURATION.md for details on available keys.
    """

    @classmethod
    def config_points(cls):
        # fmt: off
        return [
            # EuroPi revision -- this is currently unused, but reserved for future expansion
            configuration.choice(
                name="EUROPI_MODEL",
                choices = ["europi"],
                default="europi",
            ),

            # CPU & board settings
            configuration.choice(
                name="PICO_MODEL",
                choices=[
                    MODEL_PICO,
                    MODEL_PICO_W,
                    MODEL_PICO_H,
                    MODEL_PICO_2,
                    MODEL_PICO_2W,
                ],
                default=MODEL_PICO,
            ),
            configuration.choice(
                name="CPU_FREQ",
                choices=[
                    DEFAULT_FREQ,
                    OVERCLOCKED_FREQ,
                    UNDERCLOCKED_FREQ,
                ],
                default=OVERCLOCKED_FREQ,
            ),

            # Display settings
            configuration.boolean(
                name="ROTATE_DISPLAY",
                default=False,
            ),
            configuration.integer(
                name="DISPLAY_CONTRAST",
                minimum=0,
                maximum=255,
                default=255,
            ),
            configuration.integer(
                name="DISPLAY_WIDTH",
                minimum=8,
                maximum=1024,
                default=128,
                danger=True,
            ),
            configuration.integer(
                name="DISPLAY_HEIGHT",
                minimum=8,
                maximum=1024,
                default=32,
                danger=True,
            ),
            configuration.choice(
                name="DISPLAY_SDA",
                choices=[0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 26],
                default=0,
                danger=True,
            ),
            configuration.choice(
                name="DISPLAY_SCL",
                choices=[1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 27],
                default=1,
                danger=True,
            ),
            configuration.choice(
                name="DISPLAY_CHANNEL",
                choices=[0, 1],
                default=0,
                danger=True,
            ),
            configuration.choice(
                name="DISPLAY_FREQUENCY",
                choices=[
                    100000,  # 100k (Sm)
                    400000,  # 400k (Fm)
                    1000000, # 1M   (Fm+)
                    1700000, # 1.7M (Hs)
                    3400000, # 3.4M (Hs)
                    5000000, # 5M   (UFm)
                ],
                default=400000,
                danger=True,
            ),

            # External I2C connection (header between Pico & power connector)
            configuration.choice(
                name="EXTERNAL_I2C_SDA",
                choices=[0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 26],
                default=2,
            ),
            configuration.choice(
                name="EXTERNAL_I2C_SCL",
                choices=[1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 27],
                default=3,
            ),
            configuration.choice(
                name="EXTERNAL_I2C_CHANNEL",
                choices=[0, 1],
                default=1,
            ),
            configuration.choice(
                name="EXTERNAL_I2C_FREQUENCY",
                choices=[
                    100000,  # 100k (Sm)
                    400000,  # 400k (Fm)
                    1000000, # 1M   (Fm+)
                    1700000, # 1.7M (Hs)
                    3400000, # 3.4M (Hs)
                    5000000, # 5M   (UFm)
                ],
                default=100000,
            ),
            configuration.choice(
                name="EXTERNAL_I2C_TIMEOUT",
                choices=[
                    # somewhat arbitrary subset of 0-5000
                    # this should cover a good range of use-cases
                    0,
                    1,
                    5,
                    10,
                    25,
                    50,
                    100,
                    250,
                    500,
                    1000,
                    2500,
                    5000,
                ],
                default=1000,
            ),

            # I/O voltage settings
            configuration.floatingPoint(
                name="MAX_OUTPUT_VOLTAGE",
                minimum=1.0,
                maximum=10.0,
                default=10.0,
            ),
            configuration.floatingPoint(
                name="MAX_INPUT_VOLTAGE",
                minimum=1.0,
                maximum=12.0,
                default=10.0,
            ),
            configuration.floatingPoint(
                name="GATE_VOLTAGE",
                minimum=1.0,
                maximum=10.0,
                default=5.0,
            ),

            # Menu settings
            configuration.boolean(
                name="MENU_AFTER_POWER_ON",
                default=False,
            ),
        ]
        # fmt: on


def load_europi_config():
    return ConfigFile.load_config(EuroPiConfig, ConfigSpec(EuroPiConfig.config_points()))



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains objects used for interactive settings menus

Menu interaction is done using a knob and a button (K2 and B2 by default):
- rotate the knob to select the menu item
- short-press to enter edit mode
- rotate knob to select an option
- short-press button to apply the new option
- long-press button to change between the 2 menu levels (if possible)

For examples of how to create and use the SettingsMenu, please refer to
- contrib/settings_menu_example.py

Additional, more complex, examples can be found in:
- contrib/euclid.py
- contrib/pams.py (this is a very complex example)
- contrib/sequential_switch.py
- tools/conf_edit.py
"""

import europi

from configuration import *
from experimental.knobs import KnobBank, LockableKnob
from framebuf import FrameBuffer, MONO_HLSB
from machine import Timer
import os
import time


# fmt: off
AIN_GRAPHICS = bytearray(b"\x00\x00|\x00|\x00d\x00d\x00g\x80a\x80\xe1\xb0\xe1\xb0\x01\xf0\x00\x00\x00\x00")
KNOB_GRAPHICS = bytearray(b"\x06\x00\x19\x80 @@ @ \x80\x10\x82\x10A @\xa0 @\x19\x80\x06\x00")

AIN_LABEL = "AIN"
KNOB_LABEL = "Knob"

AUTOSELECT_AIN = "autoselect_ain"
AUTOSELECT_KNOB = "autoselect_knob"

DANGER_GRAPHICS = bytearray(b'\x00\x00\x04\x00\n\x00\n\x00\x11\x00\x15\x00$\x80$\x80@@D@\x80 \xff\xe0')
# fmt: off


class MenuItem:
    """
    Generic class for anything we can display in the menu

    :param parent:  A MenuItem representing this item's parent, if this item is the bottom-level of a multi-level menu
    :param children:  A list of MenuItems representing this item's children, if this is the top-level of a multi-level menu
    :param is_visible:  Is this menu item visible by default?
    """

    def __init__(
        self, children: list[object] = None, parent: object = None, is_visible: bool = True
    ):
        self.menu = None
        self.parent = parent
        self.children = children
        self.is_visible = is_visible

        # Used to indicate that if this is the active menu item, the UI should re-render it
        self.ui_dirty = False

        if parent and children:
            raise Exception("Cannot specify parent and children in the same menu item")

    def short_press(self):
        """
        Handler for when the user short-presses the button

        This does nothing by default, but can be overridden by child classes
        """
        pass

    def draw(self, oled=europi.oled):
        """
        Draw the item to the screen

        You must call the screen's ``.show()`` method after calling this

        :param oled:   A Display-compatible object we draw to
        """
        self.ui_dirty = False

    def add_child(self, item):
        """
        Add a new child item to this item

        :param item:  The menu item to add as a new child
        """
        if self.children is None:
            self.children = []
        self.children.append(item)
        item.parent = self
        item.menu = self.menu

    @property
    def is_editable(self):
        return False

    @is_editable.setter
    def is_editable(self, can_edit):
        pass

    @property
    def is_visible(self):
        return self._is_visible

    @is_visible.setter
    def is_visible(self, is_visible):
        self._is_visible = is_visible


class ChoiceMenuItem(MenuItem):
    """
    A generic menu item for displaying a choice of options to the user

    :param parent:  If the menu has multiple levels, what is this item's parent control?
    :param children:  If this menu has multiple levels, whar are this item's child controls?
    :param title:  The title to display at the top of the display when this control is active
    :param prefix:  A prefix to display before the title when this control is active
    :param graphics:  A dict of values mapped to FrameBuffer or bytearray objects, representing
        12x12 MONO_HLSB graphics to display along with the keyed values
    :param labels:  A dict of values mapped to strings, representing human-readible versions of the ConfigPoint
        options
    :param is_visible:  Is this menu item visible by default?
    """

    SELECT_OPTION_Y = 16

    def __init__(
        self,
        parent: MenuItem = None,
        children: list[MenuItem] = None,
        title: str = None,
        prefix: str = None,
        graphics: dict = None,
        labels: dict = None,
        is_visible: bool = True,
    ):
        super().__init__(
            children=children,
            parent=parent,
            is_visible=is_visible
        )

        self.title = title
        self.prefix = prefix
        self.graphics = graphics
        self.labels = labels

        self._is_editable = False

    def short_press(self):
        """Toggle is_editable on a short press"""
        self.is_editable = not self.is_editable

    def draw(self, oled=europi.oled):
        """
        Draw the current item to the display object

        You MUST call the display's .show() function after calling this in order to send the buffer to the display
        hardware

        :param oled:  A Display instance (or compatible class) to render the item
        """
        super().draw(oled)

        if self.is_editable:
            display_value = self.menu.knob.choice(self.choices)
        else:
            display_value = self.default_choice

        text_left = 0
        prefix_left = 1
        prefix_right = len(self.prefix) * europi.CHAR_WIDTH
        title_left = len(self.prefix) * europi.CHAR_WIDTH + 4

        # If we're in a top-level menu the submenu is non-empty. In that case, the prefix in inverted text
        # Otherwise, the title in inverted text to indicate we're in the sub-menu
        if self.children and len(self.children) > 0:
            oled.fill_rect(prefix_left - 1, 0, prefix_right + 1, europi.CHAR_HEIGHT + 2, 1)
            oled.text(self.prefix, prefix_left, 1, 0)
            oled.text(self.title, title_left, 1, 1)
        else:
            oled.fill_rect(
                title_left - 1,
                0,
                len(self.title) * europi.CHAR_WIDTH + 2,
                europi.CHAR_HEIGHT + 2,
                1,
            )
            oled.text(self.prefix, prefix_left, 1, 1)
            oled.text(self.title, title_left, 1, 0)

        if self.graphics:
            gfx = self.graphics.get(display_value, None)
            if gfx:
                text_left = 14  # graphics are 12x12, so add 2 pixel padding
                if type(gfx) is bytearray:
                    gfx = FrameBuffer(gfx, 12, 12, MONO_HLSB)
                oled.blit(gfx, 0, self.SELECT_OPTION_Y)

        if self.labels:
            display_text = self.labels.get(display_value, str(display_value))
        else:
            display_text = str(display_value)

        if self.is_editable:
            # draw the value in inverted text
            text_width = len(display_text) * europi.CHAR_WIDTH

            oled.fill_rect(
                text_left,
                self.SELECT_OPTION_Y,
                text_left + text_width + 3,
                europi.CHAR_HEIGHT + 4,
                1,
            )
            oled.text(display_text, text_left + 1, self.SELECT_OPTION_Y + 2, 0)
        else:
            # draw the selection in normal text
            oled.text(display_text, text_left + 1, self.SELECT_OPTION_Y + 2, 1)

    @property
    def choices(self):
        raise NotImplemented("choices(self) must be implemented by the sub-class!")

    @property
    def default_choice(self):
        raise NotImplemented("default_choice(self) must be implemented by the sub-class!")

    @property
    def is_editable(self):
        return self._is_editable

    @is_editable.setter
    def is_editable(self, can_edit):
        self._is_editable = can_edit


class SettingMenuItem(ChoiceMenuItem):
    """
    A single menu item that presents a setting the user can manipulate

    The menu item is a wrapper around a ConfigPoint, and uses
    that object's values as the available selections.

    If the item has a callback function defined, it will be invoked once during initialization

    :param config_point:  The configration option this menu item controls
    :param parent:  If the menu has multiple levels, what is this item's parent control?
    :param children:  If this menu has multiple levels, whar are this item's child controls?
    :param title:  The title to display at the top of the display when this control is active
    :param prefix:  A prefix to display before the title when this control is active
    :param graphics:  A dict of values mapped to FrameBuffer or bytearray objects, representing
        12x12 MONO_HLSB graphics to display along with the keyed values
    :param labels:  A dict of values mapped to strings, representing human-readible versions of the ConfigPoint options
    :param callback:  A function to invoke when this item's value changes. Must accept
        (new_value, old_value, config_point, arg=None) as parameters
    :param callback_arg:  An optional additional argument to pass to the callback function
    :param float_resolution:  The resolution of floating-point config points
        (ignored if config_point is not a FloatConfigPoint)
    :param value_map:  An optional dict to map the underlying simple ConfigPoint values
        to more complex objects e.g. map the string "CMaj" to a Quantizer object
    :param is_visible:  Is this menu item visible by default?
    :param autoselect_knob:  If True, this item gets "Knob" as an additional choice, allowing ad-hoc selection via the knob
    :param autoselect_cv:  If True, this item gets "AIN" as an additional choice, allowing ad-hoc selection via the CV input
    """

    def __init__(
        self,
        config_point: ConfigPoint = None,
        parent: MenuItem = None,
        children: list[MenuItem] = None,
        title: str = None,
        prefix: str = None,
        graphics: dict = None,
        labels: dict = None,
        callback=lambda new_value, old_value, config_point, arg: None,
        callback_arg=None,
        float_resolution=2,
        value_map: dict = None,
        is_visible: bool = True,
        autoselect_knob: bool = False,
        autoselect_cv: bool = False,
    ):
        if title is None:
            title = config_point.name
        if prefix is None:
            prefix = ""

        super().__init__(
            parent=parent,
            children=children,
            title=title,
            prefix=prefix,
            graphics=graphics,
            labels=labels,
            is_visible=is_visible,
        )

        self.autoselect_cv = autoselect_cv
        self.autoselect_knob = autoselect_knob
        self.value_map = value_map

        # the configuration setting that we're controlling via this menu item
        # convert everything to a choice configuration; this way we can add the knob/ain options too
        if type(config_point) is FloatConfigPoint:
            self.float_resolution = float_resolution
        self.src_config = config_point
        choices = self.get_option_list()
        self.config_point = ChoiceConfigPoint(
            config_point.name,
            choices=choices,
            default=config_point.default,
            danger=config_point.danger,
        )

        self.NUM_AUTOINPUT_CHOICES = 0
        if self.autoselect_cv or self.autoselect_knob:
            if self.autoselect_cv:
                self.NUM_AUTOINPUT_CHOICES += 1
            if self.autoselect_knob:
                self.NUM_AUTOINPUT_CHOICES += 1

            if not self.graphics:
                self.graphics = {}
            self.graphics[AUTOSELECT_AIN] = AIN_GRAPHICS
            self.graphics[AUTOSELECT_KNOB] = KNOB_GRAPHICS

            if not self.labels:
                self.labels = {}
            self.labels[AUTOSELECT_AIN] = AIN_LABEL
            self.labels[AUTOSELECT_KNOB] = KNOB_LABEL

        self.callback_fn = callback
        self.callback_arg = callback_arg

        # assign the initial value without firing any callbacks
        self._value = self.config_point.default
        self._value_choice = self.config_point.default

    @property
    def choices(self):
        return self.config_point.choices

    @property
    def default_choice(self):
        return self.value_choice

    def reset_to_default(self):
        """
        Reset this item to its default value
        """
        self.choose(self.src_config.default)

    def modify_choices(self, choices=None, new_default=None):
        """
        Regenerate this item's available choices

        This is needed if we externally modify e.g. the maximum/minimum values of the underlying
        config point as a result of one option needing to be within a range determined by another.

        :param choices:  The list of new options we want to allow the user to choose from, excluding any autoselections
        :param new_default:  A value to assign to this setting if its existing value is out-of-range
        """
        if choices is None:
            choices = self.get_option_list()
        else:
            # add the autoselect items, if needed
            if self.autoselect_knob:
                choices.append(AUTOSELECT_KNOB)
            if self.autoselect_knob:
                choices.append(AUTOSELECT_AIN)

        self.config_point.choices = choices
        still_valid = self.config_point.validate(self.value)
        if not still_valid.is_valid:
            self.choose(new_default)

    def short_press(self):
        """
        Handle a short button press

        This enters edit mode, or applies the selection and
        exits edit mode
        """
        # don't allow edit-mode for strings!
        if type(self.src_config) is StringConfigPoint:
            return

        if self.is_editable:
            new_choice = self.menu.knob.choice(self.config_point.choices)
            if new_choice != self.value_choice:
                # apply the currently-selected choice if we're in edit mode
                self.choose(new_choice)
                self.ui_dirty = True
                self.menu.settings_dirty = True

        super().short_press()

    def get_option_list(self):
        """
        Get the list of options the user can choose from

        :return:  A list of choices
        """
        t = type(self.src_config)
        if t is FloatConfigPoint:
            FLOAT_RESOLUTION = 1.0 / (10**self.float_resolution)
            items = []
            x = self.src_config.minimum
            while x <= self.src_config.maximum:
                items.append(round(x, self.float_resolution))
                x += FLOAT_RESOLUTION
            items.append(round(self.src_config.maximum, self.float_resolution))
        elif t is IntegerConfigPoint:
            items = list(range(self.src_config.minimum, self.src_config.maximum + 1))
        elif t is BooleanConfigPoint:
            items = [False, True]
        elif t is ChoiceConfigPoint:
            items = list(self.src_config.choices)  # make a copy of the items so we can append the autoselect items!
        elif t is StringConfigPoint:
            # don't allow autoselect for strings; they're not editable/selectable
            return [self.src_config.default]
        else:
            raise Exception(f"Unsupported ConfigPoint type: {type(self.src_config)}")

        # Add the autoselect inputs, if needed
        if self.autoselect_knob:
            items.append(AUTOSELECT_KNOB)
        if self.autoselect_knob:
            items.append(AUTOSELECT_AIN)

        return items

    def autoselect(self, percent: float):
        """
        Called by the parent menu when the Knob/CV timer fires, automatically updating the value of this item

        :param percent:  A value 0-1 indicating the level of the knob/cv source
        """
        last_choice = len(self.config_point.choices) - self.NUM_AUTOINPUT_CHOICES
        index = int(
            percent * last_choice
        )
        if index >= last_choice:
            index = last_choice - 1

        item = self.config_point.choices[index]
        if item != self._value:
            self.ui_dirty = True
            old_value = self._value
            self._value = item
            self.callback_fn(item, old_value, self.config_point, self.callback_arg)

    def choose(self, choice):
        """
        Set the raw value of this item's ConfigPoint

        :param choice:  The value to assign to the ConfigPoint.

        :raises ValueError: if the given choice is not valid for this setting
        """
        # choose whatever string we're given
        if type(self.src_config) is StringConfigPoint:
            self.src_config.default = choice
            self.config_point.choices = [choice]
            self._value = choice
            self._value_choice = choice
            return

        # kick out early if we aren't actually choosing anything
        if choice == self.value_choice:
            return

        validation = self.config_point.validate(choice)
        if not validation.is_valid:
            raise ValueError(f"{choice} is not a valid value for {self.config_point.name}")

        old_value = self._value_choice
        self._value_choice = choice

        if old_value == AUTOSELECT_AIN:
            self.menu.unregister_autoselect_cv(self)
        elif old_value == AUTOSELECT_KNOB:
            self.menu.unregister_autoselect_knob(self)

        if self._value_choice == AUTOSELECT_AIN:
            self.menu.register_autoselect_cv(self)
        elif self._value_choice == AUTOSELECT_KNOB:
            self.menu.register_autoselect_knob(self)
        else:
            self._value = choice
            self.callback_fn(choice, old_value, self.config_point, self.callback_arg)

    def draw(self, oled=europi.oled):
        """
        Render this item to the screen

        You must call the screen's ``.show()`` method after calling this

        :param oled: The screen we're drawing to
        """
        super().draw(oled)

        # show the real value in parentheses
        if self.value_choice == AUTOSELECT_AIN or self.value_choice == AUTOSELECT_KNOB:
            oled.text(f"({self.value})", europi.OLED_WIDTH//2, self.SELECT_OPTION_Y, 1)

        # add a ! to the lower-right corner to indicate a potentially
        # volatile, edit-at-your-own-risk item
        if self.config_point.danger:
            fb = FrameBuffer(DANGER_GRAPHICS, 12, 12, MONO_HLSB)
            oled.blit(fb, europi.OLED_WIDTH - 12, europi.OLED_HEIGHT - 12)

        if type(self.src_config) is StringConfigPoint:
            oled.text(
                "R/O",
                europi.OLED_WIDTH - 3 * europi.CHAR_WIDTH,
                europi.OLED_HEIGHT - europi.CHAR_HEIGHT,
                1
            )

    @property
    def value_choice(self):
        """The value the user has chosen from the menu"""
        return self._value_choice

    @property
    def value(self):
        """
        Get the raw value of this item's ConfigPoint

        You should use .mapped_value if you have assigned a value_map to the constructor
        """
        return self._value

    @property
    def mapped_value(self):
        """
        Get the value of this item mapped by the value_map

        If value_map was not set by the constructor, this property returns the same
        thing as .value
        """
        if self.value_map:
            return self.value_map[self._value]
        return self._value


class ActionMenuItem(ChoiceMenuItem):
    """
    A menu item that just invokes a callback function when selected.

    This class is similar to the SettingMenuItem, but doesn't wrap a ConfigPoint; it just has
    options and fires the callback when you choose one

    :param actions:  The list of choices the user can pick from. e.g. ["Cancel", "Ok"]
    :param callback:  The function to call when the user invokes the action. The selected item from choices is passed as the first parameter
    :param callback_arg:  The second parameter passed to the callback
    :param parent:  If the menu has multiple levels, what is this item's parent control?
    :param children:  If this menu has multiple levels, whar are this item's child controls?
    :param title:  The title to display at the top of the display when this control is active
    :param prefix:  A prefix to display before the title when this control is active
    :param graphics:  A dict of values mapped to FrameBuffer or bytearray objects, representing
        12x12 MONO_HLSB graphics to display along with the keyed values
    :param labels:  A dict of values mapped to strings, representing human-readible versions of the ConfigPoint options
    :param is_visible:  Is this menu item visible by default?
    """

    def __init__(
        self,
        actions: list[object],
        callback = lambda x: None,
        callback_arg: object = None,
        parent: MenuItem = None,
        children: list[MenuItem] = None,
        title: str = None,
        prefix: str = None,
        graphics: dict = None,
        labels: dict = None,
        is_visible: bool = True,
    ):
        super().__init__(
            parent=parent,
            children=children,
            title=title,
            prefix=prefix,
            graphics=graphics,
            labels=labels,
            is_visible=is_visible,
        )

        self.actions = actions
        self.callback = callback
        self.callback_arg = callback_arg
        self.title = title
        self.prefix = prefix

        self._is_editable = False

    @property
    def choices(self):
        return self.actions

    @property
    def default_choice(self):
        return self.choices[0]

    def short_press(self):
        if self.is_editable:
            # fire the callback if we're exiting edit-mode
            choice = self.menu.knob.choice(self.choices)
            self.callback(choice, self.callback_arg)

        super().short_press()



class SettingsMenu:
    """
    A menu-based GUI for any EuroPi script.

    This class is assumed to be the main interaction method for the program.

    Long/short press callbacks are invoked inside the handler for the falling edge of the button. It is recommended
    to avoid any lengthy operations inside these callbacks, as they may prevent other interrupts from being
    handled properly.

    :param menu_items:  A list of MenuItem objects representing the top-level of the menu
    :param navigation_button:  The button the user presses to interact with the menu
    :param navigation_knob:  The knob the user turns to scroll through the menu. This may be an
        experimental.knobs.KnobBank with 3 menu levels called "main_menu", "submenu" and "choice", or a raw knob like europi.k2
    :param short_press_cb:  An optional callback function to invoke when the user interacts with a short-press of the button
    :param long_press_cb:  An optional callback function to invoke when the user interacts with a long-press of the button
    :param autoselect_knob:  A knob that the user can turn to select items without needing to menu-dive
    :param autoselect_cv:  An analogue input the user can use to select items with CV
    """

    # Treat a long press as anything more than 500ms
    LONG_PRESS_MS = 500

    def __init__(
        self,
        menu_items: list = None,
        navigation_button: europi.Button = europi.b2,
        navigation_knob: europi.Knob = europi.k2,
        short_press_cb=lambda: None,
        long_press_cb=lambda: None,
        autoselect_knob: europi.Knob = europi.k1,
        autoselect_cv: europi.AnalogueInput = europi.ain,
    ):
        self._knob = navigation_knob
        self.button = navigation_button

        self._ui_dirty = True

        self.short_press_cb = short_press_cb
        self.long_press_cb = long_press_cb

        self.button.handler(self.on_button_press)
        self.button.handler_falling(self.on_button_release)

        self.items = []
        if menu_items:
            for item in menu_items:
                self.items.append(item)

        self.active_items = self.items
        self.active_item = self.knob.choice(self.items)

        self.button_down_at = time.ticks_ms()

        # Indicates to the application that we need to save the settings to disk
        self.settings_dirty = False

        # Iterate through the menu and get all of the config points
        self.config_points_by_name = {}
        self.menu_items_by_name = {}
        for item in self.items:
            item.menu = self

            if type(item) is SettingMenuItem:
                self.config_points_by_name[item.config_point.name] = item.config_point
                self.menu_items_by_name[item.config_point.name] = item

            if item.children:
                for c in item.children:
                    c.menu = self

                    if type(c) is SettingMenuItem:
                        self.config_points_by_name[c.config_point.name] = c.config_point
                        self.menu_items_by_name[c.config_point.name] = c

        # set up a timer for menu items that choose automatically based on the alternate knob or ain
        self.autoselect_cv = autoselect_cv
        self.autoselect_knob = autoselect_knob
        self.autoselect_timer = Timer()
        self.autoselect_cv_items = []
        self.autoselect_knob_items = []

    @property
    def knob(self):
        """Get the navigation knob that controls this menu"""
        if type(self._knob) is KnobBank:
            return self._knob.current
        else:
            return self._knob

    def get_config_points(self):
        """
        Get the config points for the menu so we can load/save them as needed
        """
        return list(self.config_points_by_name.values())

    def load_defaults(self, settings_file):
        """
        Load the initial settings from the file

        :param settings_file:  The path to a JSON file where the user's settings are saved
        """
        failed_key_counts = {}

        # because we may have a situation where, via callbacks, some settings' options are dynamically
        # modified, we need to load iteratively
        json_data = load_json_file(settings_file)
        keys = list(json_data.keys())
        max_tries = len(keys)
        while len(keys) > 0:
            k = keys[0]
            keys.pop(0)
            if k in self.menu_items_by_name:
                try:
                    self.menu_items_by_name[k].choose(json_data[k])
                except ValueError as err:
                    if k not in failed_key_counts:
                        failed_key_counts[k] = 1
                    else:
                        failed_key_counts[k] += 1
                    # Bump this key to the back and try again; a prior key might fix the problem
                    if failed_key_counts[k] < max_tries:
                        keys.append(k)
                    else:
                        raise

    def save(self, settings_file):
        """
        Save the current settings to the specified file

        :param settings_file: The path to the JSON file to generate
        """
        data = {}
        for item in self.menu_items_by_name.values():
            data[item.config_point.name] = item.value_choice
        try:
            # ensure the /config directory exists
            # save_to_file won't create it for us!
            os.mkdir("config")
        except OSError:
            pass
        ConfigFile.save_to_file(settings_file, data)
        self.settings_dirty = False

    def on_button_press(self):
        """Handler for the rising edge of the button signal"""
        self.button_down_at = time.ticks_ms()
        self._ui_dirty = True

    def on_button_release(self):
        """Handler for the falling edge of the button signal"""
        self._ui_dirty = True
        if time.ticks_diff(time.ticks_ms(), self.button_down_at) >= self.LONG_PRESS_MS:
            self.long_press()
        else:
            self.short_press()

    def short_press(self):
        """
        Handle a short button press

        This enters edit mode, or applies the selection and
        exits edit mode
        """
        self.active_item.short_press()

        # Cycle the knob bank, if necessary
        if type(self._knob) is KnobBank:
            if self.active_item.is_editable:
                self._knob.set_current("choice")
                if issubclass(type(self.active_item), SettingMenuItem):
                    # lock the knob to our current value
                    self._knob.current.change_lock_value(
                        self.active_item.choices.index(self.active_item.value_choice) /
                            len(self.active_item.choices)
                    )
            elif self.active_item.children and len(self.active_item.children) > 0:
                self._knob.set_current("main_menu")
            else:
                self._knob.set_current("submenu")

        self.short_press_cb()

    def long_press(self):
        """
        Handle a long button press

        This changes between the two menu levels (if possible)
        """
        # exit editable mode when we change menu levels
        self.active_item.is_editable = False

        # we're in the top-level menu; go to the submenu if it exists
        if self.active_items == self.items:
            if self.active_item.children:
                self.active_items = self.active_item.children
                if type(self.knob) is KnobBank:
                    self.knob.set_current("submenu")
        else:
            self.active_items = self.items
            if type(self.knob) is KnobBank:
                self.knob.set_current("main_menu")

        self.long_press_cb()

    def draw(self, oled=europi.oled):
        """
        Draw the menu to the given display

        You should call the display's .fill(0) function before calling this in order to clear the screen. Otherwise
        the menu item will be drawn on top of whatever is on the screen right now. (In some cases this may be the
        desired result, but when in doubt, call oled.fill(0) first).

        You MUST call the display's .show() function after calling this in order to send the buffer to the display
        hardware

        :param oled:  The display object to draw to
        """
        if not self.active_item.is_editable:
            self.active_item = self.knob.choice(self.visible_items)
        self.active_item.draw(oled)
        self._ui_dirty = False

    def register_autoselect_cv(self, menu_item: SettingMenuItem):
        """
        Connects a menu item to this menu's CV input

        :param menu_item:  The item that wants to subscribe to the CV input
        """
        if len(self.autoselect_cv_items) == 0 and len(self.autoselect_knob_items) == 0:
            self.autoselect_timer.init(freq=10, mode=Timer.PERIODIC, callback=self.do_autoselect)
        self.autoselect_cv_items.append(menu_item)

    def register_autoselect_knob(self, menu_item: SettingMenuItem):
        """
        Connects a menu item to this menu's knob input

        :param menu_item:  The item that wants to subscribe to the knob input
        """
        if len(self.autoselect_cv_items) == 0 and len(self.autoselect_knob_items) == 0:
            self.autoselect_timer.init(freq=10, mode=Timer.PERIODIC, callback=self.do_autoselect)
        self.autoselect_knob_items.append(menu_item)

    def unregister_autoselect_cv(self, menu_item: SettingMenuItem):
        """
        Disconnects a menu item to this menu's CV input

        :param menu_item:  The item that wants to unsubscribe from the CV input
        """
        self.autoselect_cv_items.remove(menu_item)
        if len(self.autoselect_cv_items) == 0 and len(self.autoselect_knob_items) == 0:
            self.autoselect_timer.deinit()

    def unregister_autoselect_knob(self, menu_item: SettingMenuItem):
        """
        Disconnects a menu item to this menu's knob input

        :param menu_item:  The item that wants to unsubscribe from the knob input
        """
        self.autoselect_knob_items.remove(menu_item)
        if len(self.autoselect_cv_items) == 0 and len(self.autoselect_knob_items) == 0:
            self.autoselect_timer.deinit()

    def do_autoselect(self, timer):
        """
        Callback function for the autoselection timer

        Reads from ain and/or the autoselect knob and applies that choice to all subscribed menu items

        :param timer: The timer instance that fired this callback
        """
        if len(self.autoselect_cv_items) > 0:
            ain_percent = self.autoselect_cv.percent()
            for item in self.autoselect_cv_items:
                item.autoselect(ain_percent)

        if len(self.autoselect_knob_items) > 0:
            knob_percent = self.autoselect_knob.percent()
            for item in self.autoselect_knob_items:
                item.autoselect(knob_percent)

    @property
    def ui_dirty(self):
        """
        Is the UI currently dirty and needs re-drawing?

        This will be true if the user has pressed the button or rotated the knob sufficiently
        to change the active item
        """
        return self._ui_dirty or self.active_item.ui_dirty or self.active_item != self.knob.choice(self.visible_items)

    @property
    def visible_items(self):
        """
        Get the set of visible menu items for the current state of the menu

        Menu items can be shown/hidden by setting their is_visible property. Normally this should be done in
        a value-change callback of a menu item to show/hide dependent other items.
        """
        items = []
        for item in self.active_items:
            if item.is_visible:
                items.append(item)
        return items



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Interface for using an NTP server as a realtime clock source

This will only work if

#. we're using a wifi-supported Pico model
#. we have a valid network configuration
#. there's an accessible NTP server on the network (important if we're in AP mode)
"""
from experimental.clocks.clock_source import ExternalClockSource
from experimental.experimental_config import load_experimental_config

import utime


class NtpError(Exception):
    """Custom NTP-related errors"""

    def __init__(self, message):
        super().__init__(message)


class NtpClock(ExternalClockSource):
    """
    Realtime clock source that uses an external NTP server

    Requires a valid network connection on a Pico W or Pico 2 W
    """

    def __init__(self):
        super().__init__()
        cfg = load_experimental_config()

        try:
            import ntptime
        except ImportError as err:
            raise NtpError(f"Failed to load NTP dependency: {err}")

        try:
            # set the clock to UTC from the default NTP source
            ntptime.settime()
        except Exception as err:
            raise NtpError(f"Failed to initialize NTP clock: {err}")

    def set_datetime(self, datetme):
        """
        NTP server is read-only, so setting the time is nonsical.

        This function won't do anything, but don't raise an error.
        """
        pass

    def datetime(self):
        """
        Get the latest time from our NTP source and return it

        :return: a tuple of the form tuple of the form (0-year, 1-month, 2-day, 3-hour, 4-minutes, 5-seconds, 6-weekday, 7-yearday)
        """
        return utime.gmtime()



# Copyright 2023 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Shared classes for quantization support

@author Chris Iverach-Brereton <ve4cib@gmail.com>
@year   2023
"""

from europi import experimental_config, MAX_OUTPUT_VOLTAGE

## 1.0V/O is the Eurorack/Moog standard, but Buchla uses 1.2V/O
VOLTS_PER_OCTAVE = experimental_config.VOLTS_PER_OCTAVE

## Standard western music scale has 12 semitones per octave
SEMITONES_PER_OCTAVE = 12

## How many volts per semitone
VOLTS_PER_SEMITONE = float(VOLTS_PER_OCTAVE) / float(SEMITONES_PER_OCTAVE)

## Labels for the 12 semitones (using sharps, not flats)
SEMITONE_LABELS = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]


class Quantizer:
    """Represents a set of semitones we can quantize input voltages to

    By default this represents a chromatic scale, with all notes enabled.  Notes can be changed
    by setting scale[n] = True/False, where n is the index of the semitone to toggle

    Implements __get_item__ and __set_item__ so you can use Quantizer like an array to set notes on/off

    :param notes:  A boolean array of length SEMITONES_PER_OCTAVE indicating what semitones
        are enabled (True) or disabled (False). If None, all notes are enabled. If not-none,
        the provided array is copied into this instance.
    :param name:  The human-readable name for this scale. The name can be displayed on the
        module's screen by some scripts.

    :raises ValueError: if len(notes) is not equal to SEMITONES_PER_OCTAVE
    """

    def __init__(self, notes=None, name=""):
        if notes is None:
            self.notes = [True] * SEMITONES_PER_OCTAVE
        else:
            if len(notes) != SEMITONES_PER_OCTAVE:
                raise ValueError(
                    f"Wrong size for notes array: {len(notes)} but expected {SEMITONES_PER_OCTAVE}"
                )

            self.notes = [n for n in notes]

        self.name = name

    def __getitem__(self, n):
        return self.notes[n % len(self.notes)]

    def __setitem__(self, n, value):
        self.notes[n % len(self.notes)] = value

    def __len__(self):
        return len(self.notes)

    def __str__(self):
        if self.name:
            return self.name
        else:
            return "".join(["1" if self.notes[i] else "0" for i in range(len(self.notes))])

    def quantize(self, analog_in, root=0):
        """Take an analog input voltage and round it to the nearest note on our scale

        :param analog_in:  The input voltage to quantize, as a float
        :param root:       An integer in the range [0, 12) indicating the number of semitones up
            to transpose the quantized scale

        :return: A tuple of the form (voltage, note) where voltage is the raw voltage to output,
            and note is a value from 0-11 indicating the semitone
        """
        # Make sure we've got at least 1 note in the scale enabled, otherwise we might have infinite loop problems
        # If we have nothing to quantize to, just output zero for both outputs
        if not (True in self.notes):
            return (0, 0)

        # offset the input voltage by the root to transpose down
        analog_in = analog_in - VOLTS_PER_SEMITONE * root

        # first get the closest chromatic voltage to the input
        nearest_chromatic_volt = round(analog_in / VOLTS_PER_SEMITONE) * VOLTS_PER_SEMITONE

        # then convert that to a 0-12 value indicating the nearest semitone
        base_volts = int(nearest_chromatic_volt)
        nearest_semitone = (nearest_chromatic_volt - base_volts) / VOLTS_PER_SEMITONE

        # go through our scale and determine the nearest on-scale note
        nearest_on_scale = 0
        best_delta = 255
        for note in range(len(self.notes)):
            if self.notes[note]:
                delta = abs(nearest_semitone - note)
                if delta < best_delta:
                    nearest_on_scale = note
                    best_delta = delta

        # re-apply the root to transpose back up
        volts = base_volts + nearest_on_scale * VOLTS_PER_SEMITONE + root * VOLTS_PER_SEMITONE

        # If the calculated voltage is above what we can actually output, move down in semitones until we get
        # to something on the scale
        # Author's Note:
        #  The likeliest way for this to trigger is if MAX_OUTPUT_VOLTAGE is set significantly lower than
        #  MAX_INPUT_VOLTAGE (e.g. 10V in, 5V out) and/or @root is set very high and @analog_in is close
        #  to MAX_OUTPUT_VOLTAGE
        highest_volts = volts
        highest_note = nearest_on_scale
        while volts > MAX_OUTPUT_VOLTAGE:
            highest_volts -= VOLTS_PER_SEMITONE
            highest_note = (highest_note - 1) % len(self.notes)
            if self.notes[highest_note]:
                volts = highest_volts
                nearest_on_scale = highest_note

        return (volts, nearest_on_scale)


class CommonScales:
    """A collection of common scales that can be used in other scripts to support quantization

    The Major135[6|7] and Minor135[6|7] scales are inspired by the Doepfer A-156 quantizer
    """

    # fmt: off
    #                         C      C#     D      D#     E      F      F#     G      G#     A      A#     B
    Chromatic    = Quantizer([True,  True,  True,  True,  True,  True,  True,  True,  True,  True,  True,  True],  "Chromatic")
    NatMajor     = Quantizer([True,  False, True,  False, True,  True,  False, True,  False, True,  False, True],  "Nat Major")
    HarMajor     = Quantizer([True,  False, True,  False, True,  True,  False, True,  True,  False, True,  False], "Har Major")
    Major135     = Quantizer([True,  False, False, False, True,  False, False, True,  False, False, False, False], "Maj 135")
    Major1356    = Quantizer([True,  False, False, False, True,  False, False, True,  False, True,  False, False], "Maj 1356")
    Major1357    = Quantizer([True,  False, False, False, True,  False, False, True,  False, False, False, True],  "Maj 1357")

    NatMinor     = Quantizer([True,  False, True,  True,  False, True,  False, True,  True,  False, True,  False], "Nat Minor")
    HarMinor     = Quantizer([True,  False, True,  True,  False, True,  False, True,  True,  False, False, True],  "Har Minor")
    Minor135     = Quantizer([True,  False, False, True,  False, False, False, True,  False, False, False, False], "Min 135")
    Minor1356    = Quantizer([True,  False, False, True,  False, False, False, True,  True,  False, False, False], "Min 1356")
    Minor1357    = Quantizer([True,  False, False, True,  False, False, False, True,  False, False, True,  False], "Min 1357")

    MajorBlues   = Quantizer([True,  False, True,  True,  True,  False, False, True,  False, True,  False, False], "Maj Blues")
    MinorBlues   = Quantizer([True,  False, False, True,  False, True,  True,  True,  False, False, True,  False], "Min Blues")

    WholeTone    = Quantizer([True,  False, True,  False, True,  False, True,  False, True,  False, True,  False], "Whole")
    Pentatonic   = Quantizer([True,  False, True,  False, True,  False, False, True,  False, True,  False, False], "Penta")
    Dominant7    = Quantizer([True,  False, False, False, True,  False, False, True,  False, False, True,  False], "Dom7")
    # fmt: on


class Intervals:
    """A collection of musical intervals as indices for the Quantizer __get_item__ and __set_item__ methods

    For example, if you have a Quantizer and want to enable the Perfect4th you can simply use

    .. code-block:: python

        q = Quantizer(...)
        q[Intervals.P4] = True

    This class uses the short notation for intervals.  See https://en.wikipedia.org/wiki/Interval_(music)
    for more information.

    The abridged summary:
    * single case-sensitive letter for the interval type
    * a number from 1-8 indicating the size of the interval
    * P = perfect
    * M = major
    * m = minor
    * d = diminished
    * A = augmented
    * three exceptions: S = semitone (equivalent to m2/A1), T = Tone (equivalent to M2/d3), TT = Tritone (equivalent to d5/A4)
    """

    # fmt: off
    P1 =  0
    d2 =  0
    m2 =  1
    A1 =  1
    S  =  1
    M2 =  2
    d3 =  2
    T  =  2
    m3 =  3
    A2 =  3
    M3 =  4
    d4 =  4
    P4 =  5
    A3 =  5
    d5 =  6
    A4 =  6
    TT =  6
    P5 =  7
    d6 =  7
    m6 =  8
    A5 =  8
    M6 =  9
    d7 =  9
    m7 = 10
    A6 = 10
    M7 = 11
    d8 = 11
    P8 = 12
    A7 = 12
    # fmt: on



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Support functions to allow using a bytearray as a bit array.

Micropython doesn't appear to have a native bitarray implementation, so this
module serves as a loose framework on top of the bytearray object to allow
easier bit-level access.
"""


def make_bit_array(length):
    """Create a bit array that contains at least length bits

    The resulting byte array will have a length rounded up to the
    next byte if length is not divisible by 8

    :param length:  The number of bits in the array

    :return: A bytearray containing at least @length bits
    """
    # Use bitwise operations instead of integer division and modulo operations to keep things fast
    if length & 0x07:
        byte_length = (length >> 3) + 1
    else:
        byte_length = length >> 3
    return bytearray(byte_length)


def get_bit(arr, index):
    """Get the value of the bit at the nth position in a bytearray

    Bytes are stored most significant bit first, so the 8th bit of [1] comes immediately after
    the first bit of `[0]`: `[ B0b7 B0b6 B0b5 B0b4 B0b3 B0b2 B0b1 B0b0 B1b7 B1b6 ... ]`

    :param arr:    The bytearray to operate on
    :param index:  The bit index to retrieve

    :return: 0 or 1, depending on the state at position index
    """
    # Use bitwise operations instead of integer division and modulo operations to keep things fast
    byte = arr[index >> 3]
    mask = 1 << ((8 - index - 1) & 0x07)
    bit = 1 if byte & mask else 0
    return bit


def set_bit(arr, index, value):
    """Set the bit at the nth position in a bytearray

    Bytes are stored most significant bit first, so the 8th bit of [1] comes immediately after
    the first bit of `[0]`: `[ B0b7 B0b6 B0b5 B0b4 B0b3 B0b2 B0b1 B0b0 B1b7 B1b6 ... ]`

    :param arr:    The bytearray to operate on
    :param index:  The bit position within the array
    :param value:  A truthy value indicating whether the bit should be set to 1 or 0
    """
    # Use bitwise operations instead of integer division and modulo operations to keep things fast
    byte = arr[index >> 3]
    mask = 1 << ((8 - index - 1) & 0x07)
    if value:
        byte = byte | mask
    else:
        byte = byte & ~mask
    arr[index >> 3] = byte


def set_all_bits(arr, value=0):
    """Set all bits in the array to the same value

    :param arr:    The bytearray to reset
    :param value:  A truthy value indicating whether all bits should be set to 0 or 1
    """
    for i in range(len(arr)):
        if value:
            arr[i] = 0xFF
        else:
            arr[i] = 0x00



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Assorted functions to make text-based logging easier

We define 4 levels of log:
- info -- general status information, high-level execution flow
- warning -- warnings that something's unusual, but not necessarily broken
- error -- a critical error that means our module has stopped working
- debug -- developer-oriented debugging information

Each log item has a level and a tag associated with it. The tag should be
unique to each module to make tracing the source of warnings easier

Log messages are written to the console and saved to /europi_log.txt. Importing
the ``europi`` module will reset the log file.
"""

import os


def log_info(message, tag=None):
    """
    Log a simple information message.

    :param message: The message to log
    :param tag: An optional tag to use as a prefix (e.g. the module name)
    """
    if tag:
        write_log_entry(f"[INFO] [{tag}] {message}")
    else:
        write_log_entry(f"[INFO] {message}")


def log_warning(message, tag=None):
    """
    Log a warning message.

    Warnings indicate an abnormal state, but are recoverable or can be worked-around.

    :param message: The message to log
    :param tag: An optional tag to use as a prefix (e.g. the module name)
    """
    if tag:
        write_log_entry(f"[WARN] [{tag}] {message}")
    else:
        write_log_entry(f"[WARN] {message}")


def log_error(message, tag=None):
    """
    Log an error message.

    Errors are critical and may indicate a crash, missing hardware, or other
    unrecoverable errors.

    :param message: The message to log
    :param tag: An optional tag to use as a prefix (e.g. the module name)
    """
    if tag:
        write_log_entry(f"[ERR ] [{tag}] {message}")
    else:
        write_log_entry(f"[ERR ] {message}")


def log_debug(message, tag=None):
    """
    Log a debugging message.

    Debug messages are for developers and can contain very low-level, code-related
    information.

    :param message: The message to log
    :param tag: An optional tag to use as a prefix (e.g. the module name)
    """
    if tag:
        write_log_entry(f"[DBUG] [{tag}] {message}")
    else:
        write_log_entry(f"[DBUG] {message}")


def write_log_entry(log_entry: str):
    """
    Write line to the log.

    When logged, the message is written to the console and saved to /europi_log.txt

    :param log_entry:  The line of text to write to the log
    """
    print(log_entry)
    try:
        with open("/europi_log.txt", "a") as log_out:
            log_out.write(log_entry)
            log_out.write("\n")
    except Exception:
        pass


def init_log():
    """
    Initialize the log file.

    This is done automatically by europi.py when it is imported
    """
    try:
        os.remove("/europi_log.txt")
    except Exception:
        pass



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from europi import *
from europi_config import EuroPiConfig
from europi_script import EuroPiScript

from configuration import ConfigFile

from experimental.knobs import KnobBank
from experimental.settings_menu import *

from framebuf import FrameBuffer, MONO_HLSB


## Lockable knob bank for K2 to make menu navigation a little easier
#
#  Note that this does mean _sometimes_ you'll need to sweep the knob all the way left/right
#  to unlock it
k2_bank = (
    KnobBank.builder(k2)
    .with_unlocked_knob("main_menu")
    .with_locked_knob("submenu", initial_percentage_value=0)
    .with_locked_knob("choice", initial_percentage_value=0)
    .build()
)


class SectionHeader(MenuItem):
    """
    A section header to contain one or more config point editors

    :param title:  The title of this section
    :param children:  A list of SettingMenuItem instances contained in this section
    """

    def __init__(self, title, children):
        super().__init__(children=children)
        self.title = title

    def draw(self, oled):
        oled.centre_text(f"{self.title}\n\nHold B2 to edit")


class ConfigurationEditor(EuroPiScript):
    """
    A global configuration-editor menu

    Used to edit the global EuroPiConfig settings -- use at your own peril!
    """

    def __init__(self):
        super().__init__()

        # fmt: off
        boolean_labels = {
            True: "Yes",
            False: "No"
        }

        voltage_items = []
        system_items = []
        display_items = []
        i2c_items = []

        config_points = EuroPiConfig.config_points()
        for cfg in config_points:
            if type(cfg) is BooleanConfigPoint:
                labels = boolean_labels
            else:
                labels = None

            # Special case; this name can't be easily modified & still fit, so just replace it
            if cfg.name == "MENU_AFTER_POWER_ON":
                title = "boot to menu"
                prefix = "Sys"
                items = system_items
            elif "EXTERNAL_I2C" in cfg.name:
                title = cfg.name.replace("EXTERNAL_I2C", "").replace("_", ' ').lower().strip()
                prefix = "I2C"
                items = i2c_items
            elif "DISPLAY" in cfg.name:
                title = cfg.name.replace("DISPLAY", "").replace("_", ' ').lower().strip()
                prefix = "OLED"
                items = display_items
            elif "VOLTAGE" in cfg.name:
                title = cfg.name.replace("VOLTAGE", "").replace("_", ' ').lower().replace("input", "in").replace("output", "out").strip()
                prefix = "Voltage"
                items = voltage_items
            else:
                title = cfg.name.replace("_", ' ').lower().strip()
                prefix = "Sys"
                items = system_items

            items.append(
                SettingMenuItem(
                    config_point=cfg,
                    labels=labels,
                    float_resolution=1,
                    prefix=prefix,
                    title=title,
                )
            )

        self.menu = SettingsMenu(
            menu_items=[
                # Voltage properties
                SectionHeader(
                    title="Voltage",
                    children=voltage_items
                ),

                # System properties
                SectionHeader(
                    title="System",
                    children=system_items
                ),

                # Display properties
                SectionHeader(
                    title="Display/OLED",
                    children=display_items
                ),

                # External I2C properties
                SectionHeader(
                    title="I2C",
                    children=i2c_items
                ),
            ],
            navigation_knob=k2_bank,
        )
        self.menu.load_defaults(ConfigFile.config_filename(EuroPiConfig))
        # fmt: on

    def main(self):
        # fmt: off
        disk_icon = bytearray(b'\xff\xf0\x9f\xe8\x9f$\x9f"\x9f!\x9f\xe1\x80\x01\xbf\xfd\xa0\x05\xaf\xf5\xa0\x05\xaf\xf5\xa0\x05\xbf\xfd\x80\x01\xff\xff')
        # fmt: on

        icon_height = OLED_HEIGHT // 2 - 8
        text_height = OLED_HEIGHT // 2 - CHAR_HEIGHT // 2

        while True:
            if self.menu.ui_dirty:
                oled.fill(0)
                self.menu.draw()
                oled.show()

            if self.menu.settings_dirty:
                # visually indicate we're saving
                oled.fill(0)
                oled.blit(FrameBuffer(disk_icon, 16, 16, MONO_HLSB), 0, icon_height)
                oled.text("Saving...", 18, text_height, 1)
                oled.show()

                self.menu.save(ConfigFile.config_filename(EuroPiConfig))
                time.sleep(0.5)

                oled.fill(0)
                self.menu.draw()
                oled.show()


if __name__ == "__main__":
    ConfigurationEditor().main()



# Copyright 2025 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Enables managing the wireless connection on the Pico W and Pico 2 W

Wifi credentials are saved in ``config/ExperimentalConfig.json``

See https://docs.micropython.org/en/latest/library/network.WLAN.html for details on wifi implementation
"""
from europi_config import (
    load_europi_config,
    MODEL_PICO_W,
    MODEL_PICO_2W,
)
from europi_hardware import b1, b2
from experimental.experimental_config import *

from europi_log import *
import utime


class WifiError(Exception):
    """Custom exception for wifi related errors"""

    def __init__(self, message):
        super().__init__(message)


class WifiConnection:
    """
    Class to manage the wifi connection

    This class is initalized automatically by the ``europi`` module, and should not be
    needed directly from any user code.

    To access/inspect the wireless connection, use the ``europi.wireless_connection``
    instance:

    .. code-block:: python

        import europi

        if europi.wifi_connection is not None:
            print(europi.wifi_connection.ip_addr)

    ``europi.wifi_connection`` will be ``None`` if:
    * the Pico model doesn't support WiFi
    * the WiFi configuration is invalid and/or raised a ``WifiError``

    The constructor will automatically do the following:
    #. Start the wireless hardware
    #. Create the access point or connect to the designated SSID (depending on mode)
    #. Configure WebREPL's default configuration (if necessary)
    #. Start WebREPL (if necessary)

    :raises WifiError: if the model doesn't support wifi, if we fail to import the
        necessary libraries, or if the network fails to connect
    """

    def __init__(self):
        try:
            import network
        except ImportError:
            raise WifiError("Failed to import network module")

        eu_cfg = load_europi_config()
        ex_cfg = load_experimental_config()

        if eu_cfg.PICO_MODEL != MODEL_PICO_2W and eu_cfg.PICO_MODEL != MODEL_PICO_W:
            raise WifiError(f"Hardware {eu_cfg.PICO_MODEL} doesn't support wifi")

        self._ssid = ex_cfg.WIFI_SSID

        nic = network.WLAN()
        if nic.status() == network.STAT_GOT_IP:
            # see if we have a lingering connection from the previous boot
            self._nic = nic
            log_info(
                f"NIC reports we already have an IP address: {self.ip_addr}. Re-using exising connection",
                "wifi",
            )
        elif ex_cfg.WIFI_MODE == WIFI_MODE_AP:
            log_info("Starting wifi in AP mode...", "wifi")
            try:
                self.connect_ap(ex_cfg)
                log_info(f"Access point {self.ssid} is up. {self.ip_addr}", "wifi")
            except Exception as err:
                raise WifiError(f"Failed to enable AP mode: {err}")
        else:
            log_info("Starting wifi in client mode...", "wifi")
            try:
                self.connect_station(ex_cfg)
                log_info(f"Connected to {self.ssid}: {self.ip_addr}", "wifi")
            except Exception as err:
                log_error(f"Failed to connect to network {ex_cfg.WIFI_SSID}: {err}", "wifi")
                raise WifiError(f"Failed to connect to network {ex_cfg.WIFI_SSID}: {err}")

        if ex_cfg.ENABLE_WEBREPL:
            log_info("Starting WebREPL...", "wifi")
            self.setup_webrepl()
            self.start_webrepl()

    def connect_station(self, cfg):
        """
        Connect to an external wireless router as a client

        :param cfg:  The ExperimentalConfig object
        """
        import network

        ssid = cfg.WIFI_SSID
        password = cfg.WIFI_PASSWORD
        bssid = cfg.WIFI_BSSID
        if len(bssid) == 0:
            bssid = None
        if password:
            security = network.WLAN.SEC_WPA_WPA2
        else:
            security = network.WLAN.SEC_OPEN

        self._nic = network.WLAN(network.WLAN.IF_STA)
        if self._nic.active():
            self._nic.active(False)

        if not cfg.WIFI_DHCP:
            self._nic.ifconfig(
                (
                    cfg.WIFI_IP,
                    cfg.WIFI_NETMASK,
                    cfg.WIFI_GATEWAY,
                    cfg.WIFI_DNS,
                )
            )

        if not self._nic.active():
            self._nic.active(True)

        # connecting can take some time, so give it a couple of tries
        max_tries = 3
        current_try = 1
        while current_try < max_tries and current_try > 0:
            log_info(f"Connecting to {ssid}... ({current_try})", "wifi")
            self._nic.connect(
                ssid=ssid,
                key=password,
                bssid=bssid,
                security=security,
            )

            connect_timeout_ms = 15_000
            start_time = utime.ticks_ms()
            abort_wifi = False
            while (
                not abort_wifi
                and not self.is_connected
                and utime.ticks_diff(utime.ticks_ms(), start_time) <= connect_timeout_ms
            ):
                abort_wifi = b1.value() != 0 or b2.value() != 0

            if abort_wifi:
                log_info("User aborted wifi connection", "wifi")
                raise WifiError("User aborted wifi connection")
            elif self.is_connected:
                log_info("Connection established!", "wifi")
                current_try = -1
            else:
                log_warning("Timed-out waiting for connection. Will try again.", "wifi")
                current_try += 1

        if current_try > 0:
            raise WifiError(f"Failed to connect to network {ssid}")

    def connect_ap(self, cfg):
        """
        Connect in access point mode

        We can optionally set the IP address and netmask in this mode

        :param cfg:  The ExperimentalConfig object
        """
        import network

        ssid = cfg.WIFI_SSID
        password = cfg.WIFI_PASSWORD
        channel = cfg.WIFI_CHANNEL

        if password:
            security = network.WLAN.SEC_WPA_WPA2
        else:
            security = network.WLAN.SEC_OPEN

        self._nic = network.WLAN(network.WLAN.IF_AP)
        if self._nic.active():
            self._nic.active(False)

        self._nic.config(
            ssid=ssid,
            channel=channel,
            key=password,
            security=security,
        )

        if not self._nic.active():
            self._nic.active(True)

    @property
    def ip_addr(self) -> str:
        """Get our current IP address"""
        (addr, _, _, _) = self._nic.ifconfig()
        return addr

    @property
    def netmask(self) -> str:
        """Get our current netmask"""
        (_, netmask, _, _) = self._nic.ifconfig()
        return netmask

    @property
    def gateway(self) -> str:
        """Get our current gateway"""
        (_, _, gateway, _) = self._nic.ifconfig()
        return gateway

    @property
    def dns(self) -> str:
        """Get our primary DNS"""
        (_, _, _, dns) = self._nic.ifconfig()
        return dns

    @property
    def ssid(self) -> str:
        """Get the SSID of our wireless network"""
        return self._ssid

    @property
    def is_connected(self) -> bool:
        """
        Is the Pico connected to anything?

        In client mode this returns True if we're connected to an access point

        In AP mode this returns True if at least one client is connected to us
        """
        return self._nic.isconnected()

    @property
    def interface(self):
        """Get the underlying network.WLAN interface"""
        return self._nic

    def setup_webrepl(self):
        """
        Create the additional configuration files needed for WebREPL.

        Creates the following files (if they do not exist already):
        * /boot.py (empty)
        * /webrepl_cfg.py (contains default WebREPL password)

        This is done automatically by the constructor; user code should not need to
        call this method
        """

        def exists(path):
            import os

            try:
                os.stat(path)
                return True
            except OSError:
                return False

        if not exists("/boot.py"):
            log_info("Creating empty /boot.py for WebREPL...", "wifi")
            with open("/boot.py", "w") as boot_py:
                boot_py.write("\n")

        if not exists("/webrepl_cfg.py"):
            log_info("Creating default /webrepl_cfg.py for WebREPL...", "wifi")
            with open("/webrepl_cfg.py", "w") as webrepl_cfg_py:
                webrepl_cfg_py.write('PASS = "EuroPi"\n')

    def start_webrepl(self):
        """
        Start the WebREPL server

        This is called automatically by the constructor; user code should not need to call this
        """
        import webrepl

        webrepl.start()
